"""Authorization ruleset handling.

The :class:`UserScopes` class handles whether a dataset, table or field can be accessed.
The other classes in this module ease to retrieval of permission objects.
"""

from __future__ import annotations

from collections.abc import Iterable, Iterator

from schematools._utils import cached_method
from schematools.types import (
    DatasetFieldSchema,
    DatasetSchema,
    DatasetTableSchema,
    Permission,
    PermissionLevel,
    ProfileDatasetSchema,
    ProfileSchema,
    ProfileTableSchema,
)

PUBLIC_SCOPE = "OPENBAAR"
RLA_SCOPE = "FEATURE/RLA"

__all__ = ("UserScopes",)


class UserScopes:
    """A request-like object that tells what the current user may access.

    A UserScopes encapsulates the scopes on a request and the profiles that apply,
    and performs permission checks against a schema.

    All ``has_...()`` functions are used for permission checks.
    Internally, these read the schema and profile data for the authorization matrix.

    * By default, all fields can be read unless the schema defines an "auth" field.
    * The "auth" flags in schema files act as a blacklist: no access, except for some roles.
    * The "profile" rules open up certain fields, hence whitelist features.
    """

    def __init__(
        self,
        query_params: dict[str, object],
        request_scopes: Iterable[str],
        all_profiles: Iterable[ProfileSchema] | None = None,
    ):
        """Initialize the user scopes object.

        Args:
            query_params: The search query filter (e.g. request.GET).
            request_scopes: The scopes granted to a request.
                Presence of the public scope "OPENBAAR" is implied.
            all_profiles: All profiles that need to be loaded.
                If not None, this iterable is stored and converted to list
                the first time it is needed.
        """
        self._query_param_names = [param for param, value in query_params.items() if value]
        self._all_profiles = all_profiles
        self._scopes = set(request_scopes) | {PUBLIC_SCOPE}

    def __repr__(self):
        return f"<UserScopes: {self._scopes!r}>"

    def __iter__(self) -> Iterator[str]:
        return iter(self._scopes)

    def add_query_params(self, params: list[str]):
        """Tell that the request has extra (implicit) parameters that are satisfied.

        For example, the detail URL of a resource already implicitly passes the
        identifier of a resource. Hence, this parameter no longer needs to be given
        found in any additional search filters or query string.
        """
        self._query_param_names.extend(params)

    @cached_method()  # type: ignore[misc]
    def has_all_scopes(self, needed_scopes: frozenset[str]) -> bool:
        """Check whether the request has all scopes.

        This performs an AND check: all scopes should be present.
        """
        return self._scopes.issuperset(needed_scopes)

    @cached_method()  # type: ignore[misc]
    def has_any_scope(self, needed_scopes: frozenset[str]) -> bool:
        """Check whether the request grants one of the given scopes.

        This performs an OR check: having one of the scopes gives access.
        """
        return not self._scopes.isdisjoint(needed_scopes)

    def has_dataset_access(self, dataset: DatasetSchema) -> Permission:
        """Tell whether a dataset can be accessed."""
        return self._has_dataset_auth_access(dataset) or self._has_dataset_profile_access(
            dataset.id
        )

    def has_table_access(self, table: DatasetTableSchema) -> Permission:
        """Tell whether a table can be accessed, and return the permission level."""
        # When the user has an "auth" scope, they may always enter.
        # Otherwise, the user can only enter when the required profile rules are satisfied,
        # which includes mandatory filtersets.
        return self._has_table_auth_access(table) or self._has_table_profile_access(table)

    def has_table_fields_access(self, table: DatasetTableSchema) -> bool:
        """Tell whether all fields of a table can be accessed."""
        for field in table.fields:  # noqa: SIM110 (needs optimized for statement instead of all())
            if not self.has_field_access(field):
                return False

        return True

    def has_field_access(self, field: DatasetFieldSchema) -> Permission:
        """Tell whether a field may be read."""
        # Again, when a field "auth" scope is satisfied, no further checks are done.
        # Otherwise, the field + table rules are checked from the profile.
        return self._has_field_auth_access(field) or self._has_field_profile_access(field)

    def has_field_filter_access(self, field: DatasetFieldSchema) -> Permission:
        """Tell whether a field may be used in searching.

        Some fields do not allow filtering the data for privacy reasons.
        An example is filtering all buildings by owner (hence knowing their portfolio),
        while the user is still allowed to see the owner of each individual building.

        This also checks whether the field has read access (:meth:`has_field_access`).
        """
        # There is no profile option yet for profile checking here.
        # Tell whether the 'filterAuth' gives extra permission to filter.
        # As this logic is an extension of the existing read-access check,
        # there is no dataset/table-level variant of the 'filterAuth'.
        if self.has_any_scope(field.filter_auth) and (field_auth := self.has_field_access(field)):
            # The if-statement checks filterAuth first. When it's defined, it likely denies access.
            # That way the other more complex checks are not needed.
            return (
                Permission(PermissionLevel.highest, source="field.filter_auth")
                if field.filter_auth
                else field_auth  # Return original permission if there is no filterAuth
            )
        else:
            return Permission.none

    def _has_dataset_auth_access(self, dataset: DatasetSchema) -> Permission:
        """Tell whether the 'auth' rules give access to the dataset."""
        if self.has_any_scope(dataset.auth):
            return Permission(PermissionLevel.highest, source="dataset.auth")
        else:
            return Permission.none

    def _has_table_auth_access(self, table: DatasetTableSchema) -> Permission:
        """Tell whether the 'auth' rules give access to the table."""
        if self.has_any_scope(table.auth) and self.has_any_scope(table.dataset.auth):
            return Permission(
                PermissionLevel.highest, source="table.auth" if table.auth else "dataset.auth"
            )
        else:
            return Permission.none

    def _has_field_auth_access(self, field: DatasetFieldSchema) -> Permission:
        """Tell whether the 'auth' rules give access to the table."""
        if (
            self.has_any_scope(field.auth)
            and self.has_any_scope(field.table.auth)
            and self.has_any_scope(field.table.dataset.auth)
        ):
            return Permission(
                PermissionLevel.highest,
                source=(
                    "field.auth"
                    if field.auth
                    else ("table.auth" if field.table.auth else "dataset.auth")
                ),
            )
        else:
            return Permission.none

    @cached_method()
    def _has_dataset_profile_access(self, dataset_id: str) -> Permission:
        """Give the permission access level for a dataset, as defined by the profile."""
        return max(
            (
                profile_dataset.permissions
                for profile_dataset in self.get_active_profile_datasets(dataset_id)
            ),
            default=Permission.none,
        )

    @cached_method()
    def _has_table_profile_access(self, table: DatasetTableSchema) -> Permission:
        """Give the permission level for a table.

        When a dataset defines global permissions without explicitly mentioning the table,
        these permissions are "inherited" and used.
        """
        dataset_id = table.dataset.id
        table_id = table.id
        max_permission = Permission.none

        for profile_dataset in self.get_active_profile_datasets(dataset_id):
            if max_permission.level == PermissionLevel.highest:
                break

            # If a profile defines "read" on the whole dataset, without explicitly
            # mentioning a table, this means the table can also be read unconditionally.
            profile_table = profile_dataset.tables.get(table_id, None)
            if profile_table is None:
                if dataset_permission := profile_dataset.permissions:
                    max_permission = max(max_permission, dataset_permission)

            # Otherwise see if the table can be included (mandatory filters match)
            elif self._may_include_profile_table(profile_table):
                max_permission = max(max_permission, profile_table.permissions)

        # Datasets may a permission that also covers this table,
        # but tables could also define an explicit permission. See who wins.
        return max_permission

    def _has_field_profile_access(self, field: DatasetFieldSchema) -> Permission:
        """Give the permission level for a field based on a profile.

        Fields have a special case: if a specific permission is defined, use that.
        This may "limit" the actual permission. For example, the table gives "read" permission,
        but the field may state "encoded" as the level. Since a default is defined for the field,
        that's being used.
        """
        field_id = field.id
        table_id = field.table.id
        max_permission = Permission.none

        # First see if there is an explicit definition for a field:
        for profile_dataset in self.get_active_profile_datasets(field.table.dataset.id):
            if max_permission.level == PermissionLevel.highest:
                break

            # If a profile defines "read" on the whole dataset, without explicitly
            # mentioning the table, this means the table can also be read unconditionally.
            profile_table = profile_dataset.tables.get(table_id, None)
            if profile_table is None:
                if dataset_permission := profile_dataset.permissions:
                    max_permission = max(max_permission, dataset_permission)
                continue

            # See if the table can be included (mandatory filters match)
            if not self._may_include_profile_table(profile_table):
                continue

            # See if the table defines the current field
            try:
                field_permission = profile_table.fields[field_id]
            except KeyError:
                # No explicit field defined, consider table as fallback
                # Some tables define global permissions without mentioning the field.
                # These get preference over explicit field permissions from other profiles.
                # When we only find profiles that mention the field explicitly,
                # that highest level will be returned instead.
                table_permission = profile_table.permissions
                if table_permission and table_permission.level > PermissionLevel.SUBOBJECTS_ONLY:
                    max_permission = max(max_permission, table_permission)
                continue

            max_permission = max(max_permission, field_permission)

        return max_permission

    @cached_method()
    def get_active_profile_datasets(self, dataset_id: str) -> list[ProfileDatasetSchema]:
        """Find all profiles that mention a dataset and match the scopes.

        This already checks whether the mandatory user scopes are set.
        """
        if self._all_profiles is None:
            self._all_profiles = []
        elif not isinstance(self._all_profiles, list):
            self._all_profiles = list(self._all_profiles)  # perform query on demand.

        return [
            profile_dataset
            for profile in self._all_profiles
            # Profiles are only activated when:
            # - ALL scopes are matched
            # - dataset is mentioned in the profile
            if self.has_all_scopes(profile.scopes)
            and (profile_dataset := profile.datasets.get(dataset_id)) is not None
        ]

    @cached_method()
    def get_active_profile_tables(
        self, dataset_id: str, table_id: str
    ) -> list[ProfileTableSchema]:
        """Find all profiles that mention a particular table and give access.

        This already checks whether the table passes the `mandatoryFilterSets` check,
        and whether the scopes of the dataset match.
        """
        return [
            profile_table
            for profile_dataset in self.get_active_profile_datasets(dataset_id)
            # Profiles are only activated when:
            # - table is mentioned in the profile
            # - ALL scopes are matched (tested for dataset already)
            # - ONE mandatory filter is matched (if filters are required)
            if (profile_table := profile_dataset.tables.get(table_id)) is not None
            and self._may_include_profile_table(profile_table)
        ]

    def _may_include_profile_table(self, profile_table: ProfileTableSchema):
        """Check whether the table rules are applicable to the current user.

        This checks whether any of the mandatory filtersets from a ProfileTableSchema were queried.
        """
        mandatory_filtersets = profile_table.mandatory_filtersets

        # Table is OK when:
        # - there are no mandatory filter sets
        # - one of the rules is matched.
        return not mandatory_filtersets or any(
            _match_filter_rule(rule, self._query_param_names) for rule in mandatory_filtersets
        )


def _match_filter_rule(rule: Iterable[str], query_param_names: Iterable[str]) -> bool:
    """Tell whether a mandatory filter rule is matched.

    This happens when ALL required filters are present in the query string.
    """
    return all(filter_name in query_param_names for filter_name in rule)



# Generated by Django 5.2.5 on 2025-09-30 11:06
from __future__ import annotations

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0023_publisher_scope"),
    ]

    operations = [
        migrations.AlterField(
            model_name="profile",
            name="id",
            field=models.CharField(max_length=100, primary_key=True, serialize=False),
        ),
    ]



"""Schematools, a library to work with amsterdam schema files."""

from __future__ import annotations

from typing import Final

# Internal conventions
RELATION_INDICATOR: Final[str] = "_"
MAX_TABLE_NAME_LENGTH: Final[int] = 63  # limitation of PostgreSQL
TMP_TABLE_POSTFIX: Final[str] = "_new"
TABLE_INDEX_POSTFIX: Final[str] = "_idx"
DATABASE_SCHEMA_NAME_DEFAULT: Final[str] = "public"

# Defaults
DEFAULT_SCHEMA_URL: Final[str] = "https://schemas.data.amsterdam.nl/datasets/"
DEFAULT_PROFILE_URL: Final[str] = "https://schemas.data.amsterdam.nl/profiles/"
# The directory where all publisher objects are defined for amsterdam-schema
PUBLISHER_DIR: Final[str] = "publishers"
# Files that can exist in publishers directory but should be ignored by
# the FileLoaders
PUBLISHER_EXCLUDE_FILES: Final[list[str]] = ["publishers.json", "index.json"]
# The directory where all scope objects are defined for amsterdam-schema
SCOPE_DIR: Final[str] = "scopes"

# Common coordinate reference systems
CRS_WGS84: Final[str] = "EPSG:4326"  # World Geodetic System 1984, used in GPS
CRS_RD_NEW: Final[str] = "EPSG:28992"  # Amersfoort / RD New
SRID_RD_NEW: Final[int] = 28992

# Some likely used 3D coordinate reference systems:
SRID_3D: Final[list[int]] = [
    7415,  # Amersfoort / RD New + NAP height
    7423,  # ETRS89 + EVRF2007 height
    9286,  # ETRS89 + NAP height
    4979,  # WGS84 + height
    4978,  # WGS84 + geocentric height
]
# The meta-schema major versions in amsterdam-schema that this
# package version is compatible with. This means that it
# can only handle schema objects that are compliant
# under at least one of these versions.
COMPATIBLE_METASCHEMAS = [4]



# Generated by Django 3.1.12 on 2021-06-23 11:35
from __future__ import annotations

from django.db import migrations, models
from django.db.models.expressions import RawSQL

import schematools.contrib.django.validators


def set_defaults(apps, schema_editor) -> None:
    """Set non nullable field `path` to value of `id`.
    Has to use RawSql beqause F() does not support JSONfield.
    """
    Dataset = apps.get_model("datasets", "Dataset")
    Dataset.objects.all().update(path=RawSQL("schema_data ->> 'id'", ()))


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0010_use_native_jsonfield"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="dataset",
            name="url_prefix",
        ),
        migrations.AddField(
            model_name="dataset",
            name="path",
            field=models.TextField(
                null=True,
                validators=[schematools.contrib.django.validators.URLPathValidator()],
            ),
        ),
        migrations.RunPython(set_defaults),
        migrations.AlterField(
            model_name="dataset",
            name="path",
            field=models.TextField(
                unique=True,
                validators=[schematools.contrib.django.validators.URLPathValidator()],
            ),
        ),
    ]



from __future__ import annotations

from typing import Any

from django.core.management import CommandParser

from schematools.contrib.django.faker.create import create_data_for

from . import BaseDatasetCommand


class Command(BaseDatasetCommand):  # noqa: D101
    help = """Create mock data for Amsterdam schema files.

    Datasets (in DSO db) + dataset tables should already have been created,
    usually with the `import_schemas --create-tables` mgm. command.

    The mocking can be limited to tables defined in the `tables` option,
    however, this only makes sense when using a single dataset.
    """  # noqa: A003
    requires_system_checks = []

    def add_arguments(self, parser: CommandParser) -> None:  # noqa: D102
        super().add_arguments(parser)
        parser.add_argument("-s", "--size", type=int, default=50, help="Number of rows")
        parser.add_argument("--sql", action="store_true", help="Generate the sql statements.")
        parser.add_argument(
            "--start-at", type=int, default=1, help="Starting number for sequences."
        )
        parser.add_argument("--table", "-t", nargs="*", help="Names of tables.")

    def handle(self, *args: list[Any], **options: dict[str, Any]) -> None:  # noqa: D102
        datasets = self.get_datasets(options, enable_db=True, default_all=False)
        if len(datasets) > 1 and options["table"]:
            self.stdout.write("The `tables` options can only be used with one dataset.")
            return
        sql_lines = create_data_for(
            *datasets,
            start_at=options["start_at"],
            size=options["size"],
            sql=options["sql"],
            tables=options["table"],
        )
        if sql_lines:
            self.stdout.write("\n".join(sql_lines))



from __future__ import annotations

import io

from django.core.management import BaseCommand, CommandError
from django.db import connection
from django.db.utils import ProgrammingError
from psycopg import sql

from schematools.contrib.django.models import Dataset, DatasetTable
from schematools.naming import to_snake_case


class Command(BaseCommand):
    """Remove Dataset objects and optionally their underlying tables."""

    help = "Drop the specified schemas"  # noqa: A003
    requires_system_checks = []

    def add_arguments(self, parser):
        """Django hook."""
        parser.add_argument(
            "schemas",
            nargs="+",
            help="Schemas that need to be dropped",
        )
        parser.add_argument(
            "--drop-tables",
            action="store_true",
            default=False,
            help="Also drop the tables associated with the schemas",
        )

    def handle(self, *args, **options):
        """Django hook implementing command logic."""
        datasets = Dataset.objects.all()
        imported_datasets = {d.name for d in datasets}
        drop_schemas = set(options.get("schemas", []))

        impossible_schemas = [ident for ident in drop_schemas if ident not in imported_datasets]
        if impossible_schemas:
            msg = io.StringIO()
            msg.write("These schemas do not exist:\n")

            for ident in impossible_schemas:
                msg.write(f"* {ident!r}")
                snake = to_snake_case(ident)
                if snake in imported_datasets:
                    msg.write(f" (did you mean {snake!r}?)")
                msg.write("\n")

            raise CommandError(msg.getvalue())

        dataset_qs = datasets.filter(name__in=drop_schemas)
        if options["drop_tables"]:
            tables = DatasetTable.objects.filter(dataset__in=dataset_qs)
            with connection.cursor() as cursor:
                for table in tables:
                    name = table.db_table
                    try:
                        cursor.execute(
                            sql.SQL("DROP TABLE {table} CASCADE;").format(
                                table=sql.Identifier(name)
                            )
                        )
                        self.stdout.write(f"Deleted table {name}")
                    except ProgrammingError as e:
                        # Dataset(Table) descriptions and their generated tables
                        # exist independently. It is therefore possible
                        # that the table is already deleted or does not exist.
                        # We only need to attempt to clean it here.
                        self.stdout.write(f"Failed to delete table {name}. Error: {e}")
                        continue

        dataset_qs.delete()
        if options["verbosity"] > 0:
            self.stdout.write(f"Deleted the following datasets: {drop_schemas}")



from __future__ import annotations

import datetime

from sqlalchemy import text

from schematools.importer.ndjson import NDJSONImporter


def test_ndjson_import_nm(here, engine, meetbouten_schema, gebieden_schema, dbsession):
    ndjson_path = here / "files" / "data" / "metingen.ndjson"
    importer = NDJSONImporter(meetbouten_schema, engine)
    importer.generate_db_objects("metingen", truncate=True, ind_extra_index=False)
    importer.load_file(ndjson_path)
    with engine.begin() as conn:
        records = (
            conn.execute(text("SELECT * from meetbouten_metingen_v1 order by identificatie"))
            .mappings()
            .all()
        )
    assert len(records) == 4
    # A non-object relation, should just lead to _id field
    assert records[0] == {"identificatie": "173", "hoortbijmeetbout_id": "13881032"}

    with engine.begin() as conn:
        # check value from the ndjson input, should be string according to the schema
        records = (
            conn.execute(
                text(
                    "SELECT * from meetbouten_metingen_refereertaanreferentiepunten_v1 order by id"
                )
            )
            .mappings()
            .all()
        )
    # Should have a field 'id' in the n-m table
    # an extra _identificatie is not needed, this is not a composite key
    assert records[0] == {
        "id": 1,
        "metingen_id": "191",
        "refereertaanreferentiepunten_id": "10180001",
    }


def test_ndjson_import_separate_relations_target_composite(
    here, engine, meetbouten_schema, gebieden_schema, dbsession
):
    ndjson_path = here / "files" / "data" / "meetbouten_ligt_in_buurt.ndjson"
    importer = NDJSONImporter(meetbouten_schema, engine)
    # NOTE: the table ID is actually "meetbouten_ligtInBuurt".
    # This also tests that both parameters still work with the old snake-cased identifiers,
    # as get_table_by_id() still does.
    importer.generate_db_objects(
        table_id="meetbouten_ligt_in_buurt",
        truncate=True,
        ind_extra_index=False,
        limit_tables_to={"meetbouten_ligt_in_buurt"},
    )

    importer.load_file(ndjson_path, is_through_table=True)

    with engine.begin() as conn:
        records = (
            conn.execute(text("SELECT * from meetbouten_meetbouten_ligt_in_buurt_v1"))
            .mappings()
            .all()
        )
    assert records == [
        {
            "id": 1,
            "meetbouten_id": "25281132",
            "ligt_in_buurt_identificatie": "03630023753951",
            "ligt_in_buurt_volgnummer": 1,
            "ligt_in_buurt_id": "03630023753951.1",
            "begin_geldigheid": datetime.date(2015, 1, 1),
            "eind_geldigheid": None,
        },
    ]


def test_ndjson_import_separate_relations_both_composite(
    here, engine, ggwgebieden_schema, dbsession
):
    ndjson_path = here / "files" / "data" / "ggwgebieden_bestaatuitbuurten.ndjson"
    importer = NDJSONImporter(ggwgebieden_schema, engine)
    importer.generate_db_objects(
        "ggwgebieden_bestaatuitbuurten", truncate=True, ind_extra_index=False
    )

    importer.load_file(ndjson_path, is_through_table=True)
    with engine.begin() as conn:
        records = (
            conn.execute(text("SELECT * FROM ggwgebieden_ggwgebieden_bestaatuitbuurten_v1"))
            .mappings()
            .all()
        )

    assert records == [
        {
            "id": 1,
            "ggwgebieden_identificatie": "012",
            "ggwgebieden_volgnummer": 1,
            "ggwgebieden_id": "012.1",
            "bestaatuitbuurten_identificatie": "045",
            "bestaatuitbuurten_volgnummer": 2,
            "bestaatuitbuurten_id": "045.2",
        },
    ]


def test_ndjson_import_no_embedded_relation_in_data(here, engine, meetbouten_schema, dbsession):
    ndjson_path = here / "files" / "data" / "meetbouten-no-embedded-rel-data.ndjson"
    importer = NDJSONImporter(meetbouten_schema, engine)
    importer.generate_db_objects("meetbouten", truncate=True, ind_extra_index=False)
    importer.load_file(ndjson_path)

    with engine.begin() as conn:
        records = conn.execute(text("SELECT * from meetbouten_meetbouten_v1")).mappings().all()
    assert records == [
        {
            "identificatie": 1,
            "ligt_in_buurt_identificatie": None,  # Should have a field identificatie and be none
            "ligt_in_buurt_volgnummer": None,
            "ligt_in_buurt_id": None,  # The foreign key, needed by Django, should be there
            "merk_code": "12",
            "merk_omschrijving": "De meetbout",
            "geometrie": "01010000204071000000000000A028FD4066666666CEBA1D41",
        }
    ]


def test_ndjson_import_no_embedded_nm_relation_in_data(
    here, engine, ggwgebieden_schema, dbsession
):
    ndjson_path = here / "files" / "data" / "ggwgebieden-no-embedded-rel-data.ndjson"
    importer = NDJSONImporter(ggwgebieden_schema, engine)
    importer.generate_db_objects("ggwgebieden", truncate=True, ind_extra_index=False)

    importer.load_file(ndjson_path, is_through_table=True)
    with engine.begin() as conn:
        records = conn.execute(text("SELECT * FROM ggwgebieden_ggwgebieden_v1")).mappings().all()
    assert len(records) == 1

    # Through table is available, but has no content
    # (because there is no data  for it in the ndjson).
    with engine.begin() as conn:
        records = (
            conn.execute(text("SELECT * FROM ggwgebieden_ggwgebieden_bestaatuitbuurten_v1"))
            .mappings()
            .all()
        )
    assert len(records) == 0


def test_ndjson_import_jsonpath_provenance(here, engine, meetbouten_schema, dbsession):
    ndjson_path = here / "files" / "data" / "meetbouten.ndjson"
    importer = NDJSONImporter(meetbouten_schema, engine)
    importer.generate_db_objects("meetbouten", truncate=True, ind_extra_index=False)
    importer.load_file(ndjson_path)
    with engine.begin() as conn:
        records = conn.execute(text("SELECT * from meetbouten_meetbouten_v1")).mappings().all()
    assert records == [
        {
            "identificatie": 1,
            "ligt_in_buurt_identificatie": "10180001",
            "ligt_in_buurt_volgnummer": 1,
            "ligt_in_buurt_id": "10180001.1",
            "merk_code": "12",
            "merk_omschrijving": "De meetbout",
            "geometrie": "01010000204071000000000000A028FD4066666666CEBA1D41",
        }
    ]


def test_ndjson_import_nm_composite_keys(here, engine, ggwgebieden_schema, dbsession):
    ndjson_path = here / "files" / "data" / "ggwgebieden.ndjson"
    importer = NDJSONImporter(ggwgebieden_schema, engine)  # TODO: fix datetime tzinfo
    importer.generate_db_objects("ggwgebieden", truncate=True, ind_extra_index=False)
    importer.load_file(ndjson_path)

    with engine.begin() as conn:
        records = conn.execute(text("SELECT * FROM ggwgebieden_ggwgebieden_v1")).mappings().all()

    # An "id" should have been generated, concat of the composite key fields
    assert records == [
        {
            "id": "03630950000000.1",
            "identificatie": "03630950000000",
            "volgnummer": 1,
            "naam": "Centrum-West",
            "code": "DX01",
            "registratiedatum": datetime.datetime(2020, 7, 21, 13, 39, 23, 856580),  # noqa: DTZ001
            "begingeldigheid": datetime.date(2014, 2, 20),
            "eindgeldigheid": datetime.date(2019, 10, 3),
            "documentdatum": datetime.date(2017, 10, 10),
            "ligtinstadsdeel_identificatie": "03630000000018",
            "ligtinstadsdeel_volgnummer": 3,
            "ligtinstadsdeel_id": "03630000000018.3",
        }
    ]

    with engine.begin() as conn:
        records = (
            conn.execute(text("SELECT * FROM ggwgebieden_ggwgebieden_bestaatuitbuurten_v1"))
            .mappings()
            .all()
        )
    assert len(records) == 3
    # Also the temporal fields are present in the database
    columns = {
        "id",
        "ggwgebieden_id",
        "bestaatuitbuurten_id",
        "ggwgebieden_volgnummer",
        "ggwgebieden_identificatie",
        "bestaatuitbuurten_identificatie",
        "bestaatuitbuurten_volgnummer",
    }

    assert set(records[0].keys()) == columns

    # Fails:
    # assert records[0] == {
    #     "id": 1,
    #     "ggwgebieden_identificatie": "03630950000000",
    #     "ggwgebieden_volgnummer": 1,
    #     "ggwgebieden_id": "03630950000000.1",
    #     "bestaatuitbuurten_identificatie": "03630023753960",
    #     "bestaatuitbuurten_volgnummer": 1,
    #     "bestaatuitbuurten_id": "03630023753960.1",
    # }


def test_ndjson_import_nm_composite_keys_with_geldigheid(here, engine, gebieden_schema, dbsession):
    ndjson_path = here / "files" / "data" / "ggwgebieden-with-geldigheid.ndjson"
    importer = NDJSONImporter(gebieden_schema, engine)  # TODO: fix datetime tzinfo
    importer.generate_db_objects("ggwgebieden", truncate=True, ind_extra_index=False)
    importer.load_file(ndjson_path)

    with engine.begin() as conn:
        records = conn.execute(text("SELECT * from gebieden_ggwgebieden_v1")).mappings().all()
    assert records == [
        {
            # An "id" should have been generated, concat of the composite key fields:
            "id": "03630950000000.1",
            "identificatie": "03630950000000",
            "volgnummer": 1,
            "registratiedatum": datetime.datetime(2020, 7, 21, 13, 39, 23, 856580),  # noqa: DTZ001
            "naam": "Centrum-West",
            "code": "DX01",
            "begin_geldigheid": None,
            "eind_geldigheid": None,
            "documentdatum": datetime.date(2017, 10, 10),
            "documentnummer": "A12",
            "ligt_in_stadsdeel_identificatie": "03630000000018",
            "ligt_in_stadsdeel_volgnummer": 3,
            "ligt_in_stadsdeel_id": "03630000000018.3",
            "geometrie": None,
        }
    ]

    with engine.begin() as conn:
        records = (
            conn.execute(
                text("SELECT * from gebieden_ggwgebieden_bestaat_uit_buurten_v1 order by id")
            )
            .mappings()
            .all()
        )
    assert len(records) == 3
    # Also the temporal fields are present in the database
    assert records[0] == {
        "id": 1,
        "ggwgebieden_id": "03630950000000.1",
        "bestaat_uit_buurten_id": "03630023753960.1",
        "ggwgebieden_identificatie": "03630950000000",
        "ggwgebieden_volgnummer": 1,
        "bestaat_uit_buurten_identificatie": "03630023753960",
        "bestaat_uit_buurten_volgnummer": 1,
        "begin_geldigheid": datetime.date(2019, 1, 12),
        "eind_geldigheid": None,
    }


def test_ndjson_import_nm_composite_selfreferencing_keys(
    here, engine, kadastraleobjecten_schema, dbsession
):
    ndjson_path = here / "files" / "data" / "kadastraleobjecten.ndjson"
    importer = NDJSONImporter(kadastraleobjecten_schema, engine)
    importer.generate_db_objects("kadastraleobjecten", truncate=True, ind_extra_index=False)
    importer.load_file(ndjson_path)

    # An "id" should have been generated, concat of the composite key fields
    with engine.begin() as conn:
        records = (
            conn.execute(text("SELECT * from brk_kadastraleobjecten_v1 order by id"))
            .mappings()
            .all()
        )
    assert records == [
        {
            "begin_geldigheid": None,
            "eind_geldigheid": None,
            "id": "KAD.001.1",
            "identificatie": "KAD.001",
            "koopsom": None,
            "neuron_id": "10",
            "registratiedatum": None,
            "soort_cultuur_onbebouwd_code": None,
            "soort_cultuur_onbebouwd_omschrijving": None,
            "soort_grootte": {"foo": 12},
            "volgnummer": 1,
        },
        {
            "begin_geldigheid": None,
            "eind_geldigheid": None,
            "id": "KAD.002.1",
            "identificatie": "KAD.002",
            "koopsom": None,
            "neuron_id": "11",
            "registratiedatum": None,
            "soort_cultuur_onbebouwd_code": "aa",
            "soort_cultuur_onbebouwd_omschrijving": "foo",
            "soort_grootte": None,
            "volgnummer": 1,
        },
    ]

    with engine.begin() as conn:
        records = (
            conn.execute(
                text("SELECT * from brk_kadastraleobjecten_is_ontstaan_uit_kadastraalobject_v1")
            )
            .mappings()
            .all()
        )
    assert records == [
        {
            "id": 1,
            "is_ontstaan_uit_kadastraalobject_id": "KAD.002.1",
            "is_ontstaan_uit_kadastraalobject_identificatie": "KAD.002",
            "is_ontstaan_uit_kadastraalobject_volgnummer": 1,
            "kadastraleobjecten_id": "KAD.001.1",
            "kadastraleobjecten_identificatie": "KAD.001",
            "kadastraleobjecten_volgnummer": 1,
        }
    ]


def test_ndjson_import_nested_tables(here, engine, verblijfsobjecten_schema, dbsession):
    ndjson_path = here / "files" / "data" / "verblijfsobjecten.ndjson"
    importer = NDJSONImporter(verblijfsobjecten_schema, engine)
    importer.generate_db_objects("verblijfsobjecten", truncate=True, ind_extra_index=False)
    importer.load_file(ndjson_path)
    with engine.begin() as conn:
        records = (
            conn.execute(
                text(
                    "SELECT code, omschrijving, parent_id"
                    " FROM verblijfsobjecten_verblijfsobjecten_gebruiksdoel_v1"
                )
            )
            .mappings()
            .all()
        )
    assert records == [
        {"code": "1", "omschrijving": "doel 1", "parent_id": "VB.1"},
        {"code": "2", "omschrijving": "doel 2", "parent_id": "VB.1"},
    ]


def test_ndjson_import_1n(here, engine, meetbouten_schema, dbsession):
    ndjson_path = here / "files" / "data" / "meetbouten.ndjson"
    importer = NDJSONImporter(meetbouten_schema, engine)
    importer.generate_db_objects("meetbouten", truncate=True, ind_extra_index=False)
    importer.load_file(ndjson_path)
    with engine.begin() as conn:
        records = conn.execute(text("SELECT * from meetbouten_meetbouten_v1")).mappings().all()
    # The foreign key, needed by Django, should be there
    # And should have the concatenated value
    # And Should have a field identificatie
    assert records == [
        {
            "identificatie": 1,
            "ligt_in_buurt_id": "10180001.1",
            "ligt_in_buurt_identificatie": "10180001",
            "ligt_in_buurt_volgnummer": 1,
            "merk_code": "12",
            "merk_omschrijving": "De meetbout",
            "geometrie": "01010000204071000000000000A028FD4066666666CEBA1D41",
        }
    ]


def test_inactive_relation_that_are_commented_out(here, engine, stadsdelen_schema, dbsession):
    """Prove that relations that are commented out in the schema are flattened to strings."""
    ndjson_path = here / "files" / "data" / "stadsdelen.ndjson"
    importer = NDJSONImporter(stadsdelen_schema, engine)
    importer.generate_db_objects("stadsdelen", truncate=True, ind_extra_index=False)
    importer.load_file(ndjson_path)
    with engine.begin() as conn:
        records = (
            conn.execute(text("SELECT * FROM stadsdelen_stadsdelen_v1 ORDER BY id"))
            .mappings()
            .all()
        )
    # Field is stringified, because in schema the relation is 'disabled'
    assert records == [
        {"id": "1", "ligt_in_gemeente": '{"identificatie": "0363"}'},
        {"id": "2", "ligt_in_gemeente": '{"identificatie": "0364"}'},
    ]


def test_missing_fields_in_jsonpath_provenance(here, engine, woonplaatsen_schema, dbsession):
    """Prove that missing fields in jsonpath provenance fields do not crash."""
    ndjson_path = here / "files" / "data" / "woonplaatsen.ndjson"
    importer = NDJSONImporter(woonplaatsen_schema, engine)
    importer.generate_db_objects("woonplaatsen", truncate=True, ind_extra_index=False)
    importer.load_file(ndjson_path)
    with engine.begin() as conn:
        records = (
            conn.execute(text("SELECT * FROM woonplaatsen_woonplaatsen_v1 ORDER BY id"))
            .mappings()
            .all()
        )
    assert records == [
        {
            "id": "1.1",
            "status_code": 1,
            "status_omschrijving": "status met code 1",
            "heeft_dossier_id": "GV12",
        },
        {
            "id": "1.2",
            "status_code": None,
            "status_omschrijving": "status zonder omschrijving",
            "heeft_dossier_id": None,
        },
    ]


def test_ndjson_import_with_shortnames_missing_data(here, engine, hr_schema, dbsession):
    """Prove that rows with missing fields with shortnames are imported correctly."""
    ndjson_path = here / "files" / "data" / "hr_missing_nmrelation.ndjson"
    importer = NDJSONImporter(hr_schema, engine)
    importer.generate_db_objects(
        "maatschappelijkeactiviteiten", truncate=True, ind_extra_index=False
    )
    importer.load_file(ndjson_path)
    with engine.begin() as conn:
        records = conn.execute(text("SELECT * from hr_activiteiten_v1")).mappings().all()

    # shortname for heeftEenRelatieMetVerblijfsobject, not in hr_missing_nmrelation.ndjson
    assert records == [
        {
            "kvknummer": "90004213",
            "gevestigd_in_identificatie": "01002",
            "gevestigd_in_volgnummer": 3,
            "gevestigd_in_id": "01002.3",
        }
    ]


def test_ndjson_import_with_shortnames_in_schema(here, engine, hr_schema, dbsession):
    """Prove that data for schemas with shortnames for tables/fields is imported correctly."""
    ndjson_path = here / "files" / "data" / "hr.ndjson"
    importer = NDJSONImporter(hr_schema, engine)
    importer.generate_db_objects(
        "maatschappelijkeactiviteiten", truncate=True, ind_extra_index=False
    )
    importer.load_file(ndjson_path)
    with engine.begin() as conn:
        records = conn.execute(text("SELECT * from hr_activiteiten_v1")).mappings().all()
        assert records == [
            {
                "kvknummer": "90004213",
                "gevestigd_in_identificatie": "01002",
                "gevestigd_in_volgnummer": 3,
                "gevestigd_in_id": "01002.3",
            }
        ]

        records = (
            conn.execute(text("SELECT * from hr_activiteiten_sbi_maatschappelijk_v1"))
            .mappings()
            .all()
        )
        assert records == [{"parent_id": "90004213", "bronwaarde": 1130, "id": 1}]

        records = (
            conn.execute(text("SELECT * from hr_activiteiten_sbi_voor_activiteit_v1"))
            .mappings()
            .all()
        )
        # In this case, the through table does not have extra fields, because the
        # FK to the target is not a composite key.
        assert records == [
            {
                "id": 1,
                "activiteiten_id": "90004213",
                "sbi_voor_activiteit_id": "01131",
            }
        ]

        records = (
            conn.execute(text("SELECT * from hr_activiteiten_verblijfsobjecten_v1"))
            .mappings()
            .all()
        )
        assert records == [
            {
                "id": 1,
                "activiteiten_id": "90004213",
                "verblijfsobjecten_id": "01001.1",
                "verblijfsobjecten_identificatie": "01001",
                "verblijfsobjecten_volgnummer": 1,
            }
        ]


def test_ndjson_import_with_shortnames_in_identifier(here, engine, hr_schema, dbsession):
    """Prove that data for schemas with shortnames in an identifier is imported correctly."""
    ndjson_path = here / "files" / "data" / "hr_sbi_act_nr.ndjson"
    importer = NDJSONImporter(hr_schema, engine)
    importer.generate_db_objects("sbiactiviteiten", truncate=True, ind_extra_index=False)
    importer.load_file(ndjson_path)

    with engine.begin() as conn:
        records = (
            conn.execute(text("SELECT * from hr_sbiactiviteiten_v1 ORDER BY sbi_act_nr"))
            .mappings()
            .all()
        )
    assert records == [{"sbi_act_nr": "12"}, {"sbi_act_nr": "16"}]


def test_ndjson_import_with_attributes_on_relation(
    here, engine, brk_schema, verblijfsobjecten_schema, dbsession
):
    """Prove that extra attributes on a relation do not interfere with ndjson import."""
    ndjson_path = here / "files" / "data" / "brk_aantek_rechten.ndjson"
    importer = NDJSONImporter(brk_schema, engine)
    importer.generate_db_objects("aantekeningenrechten", truncate=True, ind_extra_index=False)
    importer.load_file(ndjson_path)

    with engine.begin() as conn:
        records = conn.execute(text("SELECT * from brk_aantekeningenrechten_v1")).mappings().all()
    assert records == [
        {
            "aard_code": None,
            "aard_omschrijving": None,
            "betrokken_tenaamstelling_id": None,
            "einddatum": None,
            "id": "01",
            "identificatie": "01",
            "is_gbsd_op_sdl_id": "013",
            "omschrijving": None,
            "toestandsdatum": None,
        }
    ]


def test_provenance_for_schema_field_ids_equal_to_ndjson_keys(
    here, engine, woonplaatsen_schema, dbsession
):
    """Prove that imports where the schema field is equal to the key in the imported ndjson
    data are processed correctly."""
    ndjson_path = here / "files" / "data" / "woonplaatsen.ndjson"
    importer = NDJSONImporter(woonplaatsen_schema, engine)
    importer.generate_db_objects("woonplaatsen", truncate=True, ind_extra_index=False)
    importer.load_file(ndjson_path)
    with engine.begin() as conn:
        records = (
            conn.execute(text("SELECT * FROM woonplaatsen_woonplaatsen_v1 ORDER BY id"))
            .mappings()
            .all()
        )
    assert records == [
        {
            "id": "1.1",
            "status_code": 1,
            "status_omschrijving": "status met code 1",
            "heeft_dossier_id": "GV12",
        },
        {
            "id": "1.2",
            "status_code": None,
            "status_omschrijving": "status zonder omschrijving",
            "heeft_dossier_id": None,
        },
    ]



from __future__ import annotations

import json
import re
from pathlib import Path

import pytest

from schematools import validation
from schematools.permissions import PUBLIC_SCOPE
from schematools.types import DatasetSchema
from schematools.validation import (
    PROPERTIES_INTRODUCING_BREAKING_CHANGES,
    _check_display,
    _check_maingeometry,
    _identifier_properties,
    validate_dataset,
    validate_dataset_versions_version,
    validate_schema_identifier,
    validate_table,
    validate_table_version,
)


def test_camelcase() -> None:
    for ident in ("camelCase", "camelCase100"):
        assert validation._camelcase_ident(ident) is None

    error = validation._camelcase_ident("")
    assert error is not None
    assert "empty identifier" in error

    for ident, suggest in (
        ("snake_case", "snakeCase"),
        ("camelCase_snake", "camelCaseSnake"),
        ("camel100camel", "camel100Camel"),
    ):
        error = validation._camelcase_ident(ident)
        assert error is not None
        assert error.endswith(f"suggestion: {suggest}")


def test_enum_types(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("enum_types.json")

    errors = validation.run(dataset)

    error = next(errors)
    assert error.validator_name == "enum type error"
    assert error.message == "value 'foo' in field enumInts is not an integer"

    error = next(errors)
    assert error.validator_name == "enum type error"
    assert error.message == "value 2 in field enumStrs is not a string"

    error = next(errors)
    assert error.validator_name == "enum type error"
    assert error.message == "enumFloats: enum of type number not possible"

    assert list(errors) == []


def test_id_auth(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("id_auth.json")

    errors = validation.run(dataset)

    error = next(errors)
    assert error
    assert error.validator_name == "Auth on identifier field"
    assert """auth on field 'id'""" in error.message

    assert list(errors) == []


def test_id_type(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("id_type.json")

    errors = validation.run(dataset)

    error = next(errors)
    assert error
    assert error.validator_name == "Identifier field with the wrong type"
    assert """field 'uniqid'""" in error.message
    assert """'number'""" in error.message

    assert list(errors) == []


def test_id_matches_path(here: Path, schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("stadsdelen.json")

    # No errors when id equals parent path name
    errors = validation.run(dataset, str(here / "files/datasets/stadsdelen/dataset.json"))
    assert list(errors) == []

    # Error when not equal
    errors = validation.run(dataset, str(here / "files/datasets/regios/dataset.json"))
    error = next(errors)
    assert error
    assert error.validator_name == "ID does not match file path"
    assert re.match(
        r"^Id of the dataset stadsdelen does not match the parent directory"
        r" .*/files/datasets/regios\.$",
        error.message,
    )
    assert list(errors) == []

    # Test datasets in sub directory
    dataset.__setitem__("id", "beheerkaartCbsGrid")
    errors = validation.run(dataset, str(here / "files/datasets/bierkaart/cbs_grid/dataset.json"))
    error = next(errors)
    assert error
    assert error.validator_name == "ID does not match file path"
    assert re.match(
        r"^Id of the dataset beheerkaartCbsGrid does not match the parent directory"
        r" .*/files/datasets/bierkaart/cbs_grid\.$",
        error.message,
    )
    assert list(errors) == []

    errors = validation.run(
        dataset, str(here / "files/datasets/beheerkaart/cbs_grid/dataset.json")
    )
    assert list(errors) == []

    # Test identifiers ending with a number
    dataset.__setitem__("id", "covid19")
    errors = validation.run(dataset, str(here / "files/datasets/covid19/dataset.json"))
    assert list(errors) == []


def test_crs(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("crs_validation.json")

    errors = validation.run(dataset)

    error = next(errors)
    assert error
    assert error.validator_name == "crs"
    assert """No coordinate reference system defined for field""" in error.message

    assert list(errors) == []


def test_postgres_identifier_length(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("long_ids.json")

    error = next(validation.run(dataset))
    assert error
    assert error.validator_name == "PostgreSQL identifier length"
    assert "absurdly_long" in error.message

    dataset = schema_loader.get_dataset_from_file("stadsdelen.json")
    assert list(validation.run(dataset)) == []  # no validation errors


def test_postgres_duplicate_shortnames(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("duplicate_shortnames.json")

    error = next(validation.run(dataset))
    assert error
    assert error.validator_name == "PostgreSQL duplicate shortnames"
    assert error.message == "Duplicate shortname 'sameName' found for field: 'veld1,veld2'"


def test_postgres_duplicate_abbreviated_fieldnames(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("abbreviated_fieldnames.json")

    error = next(validation.run(dataset))
    assert error
    assert error.validator_name == "PostgreSQL duplicate abbreviated fieldnames"
    assert (
        error.message
        == "Fields 'eenVestigingIsGebouwOfEenComplexGebouwenDuurzameUitoefeningActiviteitenOndernemingRechtspersoon',"
        " 'eenVestigingIsGebouwOfEenComplexGebouwenDuurzameUitoefeningActiviteitenOndernemingRechtspersoon2' share "
        "the same first 63 characters. Add a shortname."
    )


def test_postgres_duplicate_abbreviated_fieldnames_with_shortname(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("abbreviated_fieldnames_with_shortname.json")
    assert list(validation.run(dataset)) == []  # no validation errors


def test_identifier_properties(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("identifier_ref.json")
    error = next(validation.run(dataset))
    assert error
    assert "foobar" in error.message

    dataset = schema_loader.get_dataset_from_file("stadsdelen.json")
    assert list(_identifier_properties(dataset)) == []  # no validation errors


def test_main_geometry(schema_loader, gebieden_schema) -> None:
    dataset = schema_loader.get_dataset_from_file("meetbouten.json")
    assert list(_check_maingeometry(dataset)) == []

    dataset.get_table_by_id("meetbouten")["schema"]["mainGeometry"] = None
    error = next(validation.run(dataset))
    assert "'mainGeometry' is required but not defined in table" in error.message

    dataset.get_table_by_id("meetbouten")["schema"]["mainGeometry"] = "not_a_geometry"
    error = next(validation.run(dataset))
    assert "mainGeometry = 'not_a_geometry'" in error.message
    assert "Field 'not_a_geometry' does not exist" in error.message

    dataset.get_table_by_id("meetbouten")["schema"]["mainGeometry"] = "merkOmschrijving"
    error = next(validation.run(dataset))
    assert error.message == (
        "mainGeometry = 'merkOmschrijving' is not a geometry field, type = 'string'"
    )


def test_display(here: Path, schema_loader) -> None:
    schema_loader.get_dataset_from_file("gebieden.json")  # fill cache
    dataset = schema_loader.get_dataset_from_file("meetbouten.json")
    assert list(_check_display(dataset)) == []

    table = dataset.get_table_by_id("meetbouten")
    table["schema"]["display"] = "not_a_field"
    error = next(validation.run(dataset))
    assert "display = 'not_a_field'" in error.message
    assert "Field 'not_a_field' does not exist" in error.message

    table["schema"]["display"] = "merkCode"
    table["schema"]["properties"]["merkCode"]["auth"] = "some_scope"
    table.__dict__.pop("fields", None)  # clear cached property
    error = next(validation.run(dataset))
    assert "'auth' property on the display field: 'merkCode' is not allowed." in error.message


def test_rel_auth_dataset(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("rel_auth.json")
    dataset["auth"] = ["HAMMERTIME"]
    dataset["reasonsNonPublic"] = ["U can't touch this"]

    errors = list(validation.run(dataset))
    assert errors == []


def test_rel_auth_dataset_public(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("rel_auth.json")
    dataset["auth"] = [PUBLIC_SCOPE]

    errors = list(validation.run(dataset))
    assert len(errors) == 0, errors


def test_rel_auth_table(here: Path) -> None:
    with (here / "files/datasets/rel_auth.json").open() as f:
        dataset_json = json.load(f)
    table = next(t for t in dataset_json["versions"]["v1"]["tables"] if t["id"] == "base")
    table["auth"] = ["HAMMERTIME"]
    table["reasonsNonPublic"] = ["U can't touch this"]
    dataset = DatasetSchema.from_dict(dataset_json)

    errors = list(validation.run(dataset))
    assert len(errors) == 1, errors
    assert "requires scopes ['HAMMERTIME']" in str(errors[0])


def test_rel_auth_field(here: Path) -> None:
    with (here / "files/datasets/rel_auth.json").open() as f:
        dataset_json = json.load(f)
    table = next(t for t in dataset_json["versions"]["v1"]["tables"] if t["id"] == "base")
    field = table["schema"]["properties"]["stop"]
    field["auth"] = ["HAMMERTIME"]

    dataset = DatasetSchema.from_dict(dataset_json)
    errors = list(validation.run(dataset))

    assert len(errors) >= 1, errors
    assert any("requires scopes ['HAMMERTIME']" in str(e) for e in errors)


@pytest.mark.skip(reason="See comment in validator function")
def test_repetitive_naming(here: Path, schema_loader) -> None:
    dataset = schema_loader.get_dataset("repetitive")
    errors = {str(e) for e in validation.run(dataset)}

    assert errors == {
        "[repetitive identifiers] " + e
        for e in [
            "table name 'repetitiveTable' should not start with 'repetitive'",
            "field name 'repetitiveTableField' should not start with 'repetitive'",
            "field name 'repetitiveTableField' should not start with 'repetitiveTable'",
        ]
    }


def test_reasons_non_public_exists(here: Path, schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("hr_auth.json")
    errors = list(validation.run(dataset))

    # Test an error is given for the highest non-public scope
    # and only for the highest non-public scope.
    assert len(errors) == 1
    assert errors[0].message == "Non-public dataset hr should have a 'reasonsNonPublic' property."

    dataset["auth"] = [PUBLIC_SCOPE]
    errors = list(validation.run(dataset))
    assert len(errors) == 1
    assert (
        errors[0].message
        == "Non-public table sbiactiviteiten should have a 'reasonsNonPublic' property."
    )

    # Test no error is given when a reason is present
    dataset_json = json.load((here / "files/datasets/hr_auth.json").open())
    dataset_json["versions"]["v1"]["tables"][0]["reasonsNonPublic"] = [
        "5.1 1c: Bevat persoonsgegevens"
    ]
    dataset = DatasetSchema.from_dict(dataset_json)
    dataset["auth"] = [PUBLIC_SCOPE]
    errors = list(validation.run(dataset))
    assert len(errors) == 0


def test_reasons_non_public_value(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("hr_auth.json")
    dataset["reasonsNonPublic"] = ["5.1 1c: Bevat persoonsgegevens", "nader te bepalen"]
    errors = list(validation.run(dataset))

    # Test an error is given for the placeholder value in a dataset with status = beschikbaar.
    assert len(errors) == 1
    assert "not allowed in ReasonsNonPublic property of dataset hr." in errors[0].message

    # Test no error is given for the placeholder value in a dataset with status != beschikbaar.
    dataset.versions["v1"]["enableAPI"] = False
    errors = list(validation.run(dataset))
    assert len(errors) == 0

    # Test no error is given for other values of reasonsNonPublic
    dataset.versions["v1"]["enableAPI"] = True
    dataset["reasonsNonPublic"] = ["5.1 1c: Bevat persoonsgegevens"]
    errors = list(validation.run(dataset))
    assert len(errors) == 0


def test_schema_ref(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("schema_ref_validation.json")

    errors = validation.run(dataset)

    error = next(errors)
    assert error
    assert error.validator_name == "schema ref"
    assert """Incorrect `$ref` for""" in error.message

    assert list(errors) == []


def test_check_default_version(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("schema_default_version.json")

    errors = list(validation.run(dataset))
    assert len(errors) == 0

    # Prove that disabled default version gives an error (when there are multiple versions)
    dataset["defaultVersion"] = "v2"
    dataset["versions"]["v2"]["status"] = "niet_beschikbaar"
    errors = list(validation.run(dataset))
    assert len(errors) == 1
    assert "Default version v2 is not enabled." in errors[0].message


def test_check_default_version_is_experimental(schema_loader) -> None:
    """Ensure that if there is only one version and it is experimental, it does not matter that
    the default version is unavailable."""
    dataset = schema_loader.get_dataset_from_file("schema_default_version_experimental.json")

    errors = list(validation.run(dataset))
    assert len(errors) == 0


def test_production_version_tables(schema_loader) -> None:
    dataset = schema_loader.get_dataset("production_version")

    errors = list(validation.run(dataset))
    assert len(errors) == 1
    assert (
        "Dataset version (v1) cannot contain non-production table [tables/v0]" in errors[0].message
    )


def test_row_level_auth(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("brp_row_level_auth.json")
    errors = list(validation.run(dataset))
    assert len(errors) == 0


def test_row_level_auth_fail(schema_loader) -> None:
    dataset = schema_loader.get_dataset_from_file("brp_row_level_auth_fail.json")
    errors = [v.message for v in validation.run(dataset)]
    assert len(errors) == 6
    assert errors == [
        "Source verblijfplaats.afgeschermd_adres is not available in table Ingeschrevenpersonen.",
        "Target verblijfplaats.telefoon does not exist in table Ingeschrevenpersonen",
        "Source verblijfplaats.afgeschermd in table Ingeschrevenpersonen2 is not a boolean.",
        "Source verblijfplaats.afgeschermd is also a target!",
        "Target verblijfplaats.huisnummer does not define FEATURE/RLA auth.",
        "Target bestaat.niet does not exist in table Ingeschrevenpersonen2",
    ]


def test_subresource(schema_loader) -> None:
    schema_loader.get_dataset_from_file("woningbouwplannen.json")
    dataset = schema_loader.get_dataset_from_file("subresources_invalid.json")
    errors = [v.message for v in validation.run(dataset)]

    # There are two subresources defined in the schema, but only one refers to a table in the
    # same dataset. The other should give an error.
    assert len(errors) == 3
    assert errors == [
        "Subresource woningbouwplannen:woningbouwplan is not part of the same dataset as "
        "gebieden:wijken. Subresources must always be part of the same dataset.",
        "Table pleinen does not exist in dataset gebieden. Cannot use as subresource.",
        "Field ligtInDezeWijk does not exist on table gebieden:straten. Cannot use as subresource.",
    ]


def test_check_status(schema_loader) -> None:
    dataset = schema_loader.get_dataset("status")

    errors = list(validation.run(dataset))
    assert len(errors) == 1
    assert (
        "Dataset version (v0) cannot have a status of 'stable' while being a non-production version."
        in errors[0].message
    )


def test_check_enable_api(schema_loader) -> None:
    dataset = schema_loader.get_dataset("metaschemav4/enableapi")

    errors = list(validation.run(dataset))
    assert len(errors) == 1
    assert "Default version v1 is not enabled." in errors[0].message


def test_check_superseded_version(schema_loader) -> None:
    dataset = schema_loader.get_dataset("metaschemav4/supersededversion")

    errors = list(validation.run(dataset))
    assert len(errors) == 1
    assert (
        "Dataset version (v1) cannot have a status of 'superseded' without an endSupportDate."
        in errors[0].message
    )


@pytest.mark.parametrize(
    "prev,curr,errors",
    [
        # No changes
        ([{"id": "table", "$ref": "table/v1"}], [{"id": "table", "$ref": "table/v1"}], []),
        # Added table
        (
            [{"id": "table", "$ref": "table/v1"}],
            [{"id": "table", "$ref": "table/v1"}, {"id": "table2", "$ref": "table2/v1"}],
            [],
        ),
        # Removed table
        (
            [{"id": "table", "$ref": "table/v1"}, {"id": "table2", "$ref": "table2/v1"}],
            [{"id": "table", "$ref": "table/v1"}],
            ["Table table2 has been removed."],
        ),
        # Changed table version
        (
            [{"id": "table", "$ref": "table/v1"}],
            [{"id": "table", "$ref": "table/v2"}],
            [
                "Table table has changed version. Previous version: table/v1, current version: table/v2."
            ],
        ),
        # Multiple errors
        (
            [{"id": "table", "$ref": "table/v1"}, {"id": "table2", "$ref": "table2/v1"}],
            [{"id": "table", "$ref": "table/v2"}],
            [
                "Table table has changed version. Previous version: table/v1, current version: table/v2.",
                "Table table2 has been removed.",
            ],
        ),
    ],
)
def test_validate_dataset(prev, curr, errors):
    table_errors = validate_dataset(prev, curr)
    assert table_errors == errors


@pytest.mark.parametrize(
    "prev,curr,errors",
    [
        (
            "value",
            "changed_value",
            [
                "Identifier field would be changed from value to changed_value. "
                "Changing the table schema identifier is a breaking change."
            ],
        ),
        (
            ["value"],
            ["changed_value"],
            [
                "Identifier field would be changed from ['value'] to ['changed_value']. "
                "Changing the table schema identifier is a breaking change."
            ],
        ),
        (
            None,
            ["changed_value"],
            [
                "Identifier field would be changed from None to ['changed_value']. "
                "Changing the table schema identifier is a breaking change."
            ],
        ),
    ],
)
def test_validate_table_schema_identifier(prev, curr, errors):
    table_errors = validate_schema_identifier(prev, curr)
    assert table_errors == errors


@pytest.mark.parametrize(
    "prev,curr,errors",
    [
        # No changes
        ({"name": {"type": "string"}}, {"name": {"type": "string"}}, []),
        # Deleted field
        ({"name": {"type": "string"}}, {}, ["Column name would be deleted."]),
        # Changed schema ref
        (
            {
                "schema": {
                    "$ref": "https://schemas.data.amsterdam.nl/schema@v1.1.1#/definitions/schema"
                }
            },
            {
                "schema": {
                    "$ref": "https://schemas.data.amsterdam.nl/schema@v3.1.0#/definitions/schema"
                }
            },
            [],
        ),
        # Changed array item type
        (
            {"list": {"type": "array", "items": {"type": "string"}}},
            {"list": {"type": "array", "items": {"type": "integer"}}},
            ["Column list would change items.type."],
        ),
        # Changed object property.
        (
            {"object": {"type": "object", "properties": {"element": {"type": "string"}}}},
            {"object": {"type": "object", "properties": {"element": {"type": "integer"}}}},
            ["Column object would change element.type."],
        ),
        # Object with format JSON works does not raise errors.
        (
            {"object": {"type": "object", "format": "json"}},
            {"object": {"type": "object", "format": "json"}},
            [],
        ),
        # Changed object property type within an array
        (
            {
                "list": {
                    "type": "array",
                    "items": {"type": "object", "properties": {"element": {"type": "string"}}},
                }
            },
            {
                "list": {
                    "type": "array",
                    "items": {"type": "object", "properties": {"element": {"type": "integer"}}},
                }
            },
            ["Column list would change items.element.type."],
        ),
        # Changed subproperty of a property of an object
        (
            {
                "object": {
                    "type": "object",
                    "properties": {
                        "element": {
                            "type": "object",
                            "properties": {"subelement": {"type": "string"}},
                        }
                    },
                }
            },
            {
                "object": {
                    "type": "object",
                    "properties": {
                        "element": {
                            "type": "object",
                            "properties": {"subelement": {"type": "integer"}},
                        }
                    },
                }
            },
            ["Column object would change element.subelement.type."],
        ),
        # Multiple breaking changes
        (
            {
                "object": {
                    "type": "object",
                    "properties": {
                        "element": {
                            "type": "object",
                            "properties": {
                                "subelement": {"type": "string"},
                                "subelement2": {"type": "string"},
                            },
                        }
                    },
                    "description": "Original description",
                    "relation": "table1:object_1",
                },
                "id": {"type": "string"},
                "deprecated_field": {"type": "string"},
            },
            {
                "object": {
                    "type": "object",
                    "properties": {
                        "element": {
                            "type": "object",
                            "properties": {"subelement": {"type": "integer"}},
                        }
                    },
                    "description": "Description 2.0",
                    "relation": "table2:object_2",
                },
                "id": {"type": "integer"},
            },
            [
                "Column object would change relation.",
                "Column object would change element.subelement.type.",
                "Property element.subelement2 would be deleted from column object.",
                "Column id would change type.",
                "Column deprecated_field would be deleted.",
            ],
        ),
    ]
    + [
        # All properties that should stay the same.
        (
            {"property": {prop: "string"}},
            {"property": {prop: "integer"}},
            [f"Column property would change {prop}."],
        )
        for prop in PROPERTIES_INTRODUCING_BREAKING_CHANGES
    ],
)
def test_validate_table(prev, curr, errors):
    table_errors = validate_table(prev, curr)
    assert table_errors == errors


@pytest.mark.parametrize(
    "prev,curr,errors",
    [
        # No updates
        (
            {"id": "test", "version": "1.0.0", "schema": {"properties": {"field": {}}}},
            {"id": "test", "version": "1.0.0", "schema": {"properties": {"field": {}}}},
            [],
        ),
        # Add field success
        (
            {"id": "test", "version": "1.0.0", "schema": {"properties": {"field": {}}}},
            {
                "id": "test",
                "version": "1.1.0",
                "schema": {"properties": {"field": {}, "field2": {}}},
            },
            [],
        ),
        # Add field fail no new version
        (
            {"id": "test", "version": "1.0.0", "schema": {"properties": {"field": {}}}},
            {
                "id": "test",
                "version": "1.0.0",
                "schema": {"properties": {"field": {}, "field2": {}}},
            },
            ["Table 'test' added fields, expecting new version to be 1.1.0."],
        ),
        # Add field fail only patch bump
        (
            {"id": "test", "version": "1.0.0", "schema": {"properties": {"field": {}}}},
            {
                "id": "test",
                "version": "1.0.1",
                "schema": {"properties": {"field": {}, "field2": {}}},
            },
            ["Table 'test' added fields, expecting new version to be 1.1.0."],
        ),
        # Change metadata fail no new version
        (
            {
                "id": "test",
                "title": "Title 1",
                "description": "Description 1",
                "shortname": "test",
                "version": "1.0.0",
                "schema": {"properties": {"field": {}}},
            },
            {
                "id": "test",
                "title": "Title 2",
                "description": "Description 2",
                "shortname": "test2",
                "version": "1.0.0",
                "schema": {"properties": {"field": {}}},
            },
            [
                "Property 'title' on table 'test' has changed, expecting new version to be 1.0.1.",
                "Property 'description' on table 'test' has changed, expecting new version to be 1.0.1.",
                "Property 'shortname' on table 'test' has changed, expecting new version to be 1.0.1.",
            ],
        ),
        # Change metadata fail version too high
        (
            {
                "id": "test",
                "title": "Title 1",
                "description": "Description 1",
                "shortname": "test",
                "version": "1.0.0",
                "schema": {"properties": {"field": {}}},
            },
            {
                "id": "test",
                "title": "Title 2",
                "description": "Description 2",
                "shortname": "test2",
                "version": "1.1.0",
                "schema": {"properties": {"field": {}}},
            },
            [
                "Property 'title' on table 'test' has changed, expecting new version to be 1.0.1.",
                "Property 'description' on table 'test' has changed, expecting new version to be 1.0.1.",
                "Property 'shortname' on table 'test' has changed, expecting new version to be 1.0.1.",
            ],
        ),
        # Change metadata success
        (
            {
                "id": "test",
                "title": "Title 1",
                "description": "Description 1",
                "shortname": "test",
                "version": "1.0.0",
                "schema": {"properties": {"field": {}}},
            },
            {
                "id": "test",
                "title": "Title 2",
                "description": "Description 2",
                "shortname": "test2",
                "version": "1.0.1",
                "schema": {"properties": {"field": {}}},
            },
            [],
        ),
        # Change metadata on schema fail no new version
        (
            {
                "id": "test",
                "version": "1.0.0",
                "schema": {"display": "field", "properties": {"field": {}}},
            },
            {
                "id": "test",
                "version": "1.0.0",
                "schema": {"display": "field2", "properties": {"field": {}}},
            },
            [
                "Property 'schema.display' on table 'test' has changed, expecting new version to be 1.0.1."
            ],
        ),
        # Change metadata on schema success
        (
            {
                "id": "test",
                "version": "1.0.0",
                "schema": {"display": "field", "properties": {"field": {}}},
            },
            {
                "id": "test",
                "version": "1.0.1",
                "schema": {"display": "field2", "properties": {"field": {}}},
            },
            [],
        ),
        # Change metadata on field fail no new version
        (
            {
                "id": "test",
                "version": "1.0.0",
                "schema": {
                    "properties": {
                        "field": {
                            "title": "field",
                            "unit": "m2",
                            "shortname": "field",
                            "description": "Field1",
                        }
                    }
                },
            },
            {
                "id": "test",
                "version": "1.0.0",
                "schema": {
                    "properties": {
                        "field": {
                            "title": "field2",
                            "unit": "cm2",
                            "shortname": "field2",
                            "description": "Field2",
                        }
                    }
                },
            },
            [
                "Property 'title' on field 'field' in table 'test' has changed, expecting new version to be 1.0.1.",
                "Property 'description' on field 'field' in table 'test' has changed, expecting new version to be 1.0.1.",
                "Property 'shortname' on field 'field' in table 'test' has changed, expecting new version to be 1.0.1.",
                "Property 'unit' on field 'field' in table 'test' has changed, expecting new version to be 1.0.1.",
            ],
        ),
        # Change metadata on field success
        (
            {
                "id": "test",
                "version": "1.0.0",
                "schema": {
                    "properties": {
                        "field": {
                            "title": "field",
                            "unit": "m2",
                            "shortname": "field",
                            "description": "Field1",
                        }
                    }
                },
            },
            {
                "id": "test",
                "version": "1.0.1",
                "schema": {
                    "properties": {
                        "field": {
                            "title": "field2",
                            "unit": "cm2",
                            "shortname": "field2",
                            "description": "Field2",
                        }
                    }
                },
            },
            [],
        ),
        # Change major version number on table fail
        (
            {
                "id": "test",
                "version": "1.0.0",
                "schema": {
                    "properties": {
                        "field": {
                            "title": "field",
                            "unit": "m2",
                            "shortname": "field",
                            "description": "Field1",
                        }
                    }
                },
            },
            {
                "id": "test",
                "version": "2.0.0",
                "schema": {
                    "properties": {
                        "field": {
                            "title": "field2",
                            "unit": "cm2",
                            "shortname": "field2",
                            "description": "Field2",
                        }
                    }
                },
            },
            ["Table 'test' changed major version, a new file v2.json should be created."],
        ),
    ],
)
def test_validate_table_version(prev, curr, errors):
    table_errors = validate_table_version(prev, curr)
    assert table_errors == errors


@pytest.mark.parametrize(
    "prev,curr,errors",
    [
        # No changes, no fail
        (
            {"version": "1.0.0", "status": "stable", "tables": [{"id": "table1"}]},
            {"version": "1.0.0", "status": "stable", "tables": [{"id": "table1"}]},
            [],
        ),
        # Changes with version bump, no fail
        (
            {"version": "1.0.0", "status": "stable", "tables": [{"id": "table1"}]},
            {
                "version": "1.1.0",
                "status": "stable",
                "tables": [{"id": "table1"}, {"id": "table2"}],
            },
            [],
        ),
        # Changes for experimental table, no fail
        (
            {"version": "1.0.0", "status": "under_development", "tables": [{"id": "table1"}]},
            {
                "version": "1.1.0",
                "status": "under_development",
                "tables": [{"id": "table1"}, {"id": "table2"}],
            },
            [],
        ),
        # Changes without version bump, fail
        (
            {"version": "1.0.0", "status": "stable", "tables": [{"id": "table1"}]},
            {
                "version": "1.0.0",
                "status": "stable",
                "tables": [{"id": "table1"}, {"id": "table2"}],
            },
            ["Dataset 'dataset' v1 has an added table, expecting new version to be 1.1.0."],
        ),
        # Changes with too big of a minor version bump, fail
        (
            {"version": "1.0.0", "status": "stable", "tables": [{"id": "table1"}]},
            {
                "version": "1.2.0",
                "status": "stable",
                "tables": [{"id": "table1"}, {"id": "table2"}],
            },
            ["Dataset 'dataset' v1 has an added table, expecting new version to be 1.1.0."],
        ),
        # Changes with a patch version expect new minor version without patch, no fail
        (
            {"version": "1.1.1", "status": "stable", "tables": [{"id": "table1"}]},
            {
                "version": "1.2.0",
                "status": "stable",
                "tables": [{"id": "table1"}, {"id": "table2"}],
            },
            [],
        ),
        # Changes with a major version bump, fail
        (
            {"version": "1.0.0", "status": "stable", "tables": [{"id": "table1"}]},
            {
                "version": "2.0.0",
                "status": "stable",
                "tables": [{"id": "table1"}, {"id": "table2"}],
            },
            ["Dataset 'dataset' v1 has an added table, expecting new version to be 1.1.0."],
        ),
        # Changes with too little of a version bump, fail
        (
            {"version": "1.0.0", "status": "stable", "tables": [{"id": "table1"}]},
            {
                "version": "1.0.1",
                "status": "stable",
                "tables": [{"id": "table1"}, {"id": "table2"}],
            },
            ["Dataset 'dataset' v1 has an added table, expecting new version to be 1.1.0."],
        ),
    ],
)
def test_validate_dataset_versions_version(prev, curr, errors):
    dataset_version_errors = validate_dataset_versions_version("dataset", prev, curr)
    assert dataset_version_errors == errors



from __future__ import annotations

import contextlib
import json
import os
from collections.abc import Callable
from functools import cached_property
from pathlib import Path
from urllib.parse import urlparse

import requests
from more_ds.network.url import URL

from schematools import (
    DEFAULT_PROFILE_URL,
    DEFAULT_SCHEMA_URL,
    PUBLISHER_DIR,
    PUBLISHER_EXCLUDE_FILES,
    SCOPE_DIR,
)
from schematools.exceptions import (
    DatasetNotFound,
    DatasetTableNotFound,
    DuplicateProfileId,
    DuplicateScopeId,
    SchemaObjectNotFound,
)
from schematools.types import (
    DatasetSchema,
    DatasetTableSchema,
    Json,
    ProfileSchema,
    Publisher,
    Scope,
)

__all__ = (
    "get_schema_loader",
    "CachedSchemaLoader",
    "FileSystemSchemaLoader",
    "URLSchemaLoader",
    "SchemaLoader",
)


class SchemaLoader:
    """Interface that defines what a schema loader should provide."""

    def _get_dataset(self, dataset_id: str, prefetch_related: bool = False) -> DatasetSchema:
        """Gets a dataset for dataset_id."""
        raise NotImplementedError

    def _get_table(self, dataset: DatasetSchema, table_ref: str) -> DatasetTableSchema:
        """Retrieves a versioned table by reference"""
        raise NotImplementedError

    def _get_view(self, dataset: DatasetSchema, table_ref: str) -> str:
        """Retrieves a view by reference"""
        raise NotImplementedError

    def _get_dataset_path(self, dataset_id: str) -> str:
        """Find the relative path of a dataset within the location"""
        raise NotImplementedError

    def _get_all_datasets(self) -> dict[str, DatasetSchema]:
        """Gets all datasets from the schema_url location.

        The return value maps dataset paths (foo/bar) to schema's.
        """
        raise NotImplementedError

    def _get_all_publishers(self) -> dict[str, Publisher]:
        """Get all publishers from the schema location

        The return value maps pulisher ids to Publisher objects.
        """
        raise NotImplementedError

    def _get_all_scopes(self) -> dict[str, Scope]:
        """Get all scopes from the schema location

        The return value maps scope ids to Scope objects.
        """
        raise NotImplementedError

    def _get_publisher(self, publisher_id: str) -> Publisher:
        raise NotImplementedError

    def _get_scope(self, ref: str) -> Scope:
        raise NotImplementedError


class ProfileLoader:
    """Interface for loading profile objects"""

    def get_profile(self, profile_id: str) -> ProfileSchema:
        raise NotImplementedError()

    def get_all_profiles(self) -> list[ProfileSchema]:
        raise NotImplementedError()


class CachedSchemaLoader(SchemaLoader):
    """Base class for a loader that caches the results."""

    def __init__(self):
        """Initialize the cache.
        When the loader is not defined, this acts as a simple cache.
        """
        self._cache: dict[str, DatasetSchema] = {}
        self._publisher_cache: dict[str, Publisher] = {}
        self._scopes_cache: dict[str, Scope] = {}
        self._table_cache: dict[tuple[str, str], DatasetTableSchema] = {}
        self._has_all_publishers = False
        self._has_all_scopes = False
        self._has_all = False

    def __repr__(self):
        return f"{self.__class__.__name__}({self._loader!r})"

    def add_dataset(self, dataset: DatasetSchema) -> None:
        """Add a dataset to the cache."""
        self._cache[dataset.id] = dataset

    def clear(self) -> None:
        """Clear the cache."""
        self._cache.clear()
        self._table_cache.clear()

    def get_dataset_path(self, dataset_id) -> str:
        return self._get_dataset_path(dataset_id)

    def get_dataset(self, dataset_id: str, prefetch_related: bool = False) -> DatasetSchema:
        """Gets a dataset by id from the cache.

        If not available, load the dataset from the SCHEMA_URL location.
        NB. Because dataset schemas are imported into the Postgresql database
        by the DSO API, there is a chance that the dataset that is loaded from SCHEMA_URL
        differs from the definition that is in de Postgresql database.
        """
        try:
            return self._cache[dataset_id]
        except KeyError:
            dataset = self._get_dataset(dataset_id, prefetch_related=prefetch_related)
            self.add_dataset(dataset)
            return dataset

    def get_table(self, dataset: DatasetSchema, table_ref: str) -> DatasetTableSchema:
        key = (dataset.id, table_ref)
        try:
            return self._table_cache[key]
        except KeyError:
            table = self._get_table(dataset, table_ref)
            self._table_cache[key] = table
            return table

    def get_all_datasets(self) -> dict[str, DatasetSchema]:
        """Load all datasets, and fill the cache"""
        if not self._has_all:
            self._cache = {schema.id: schema for schema in self._get_all_datasets().values()}
            self._has_all = True

        return self._cache

    def get_publisher(self, publisher_id: str) -> Publisher:
        if (publisher := self._publisher_cache.get(publisher_id)) is not None:
            return publisher

        publisher = self._get_publisher(publisher_id)
        self._publisher_cache[publisher.id] = publisher
        return publisher

    def get_all_publishers(self) -> dict[str, Publisher]:
        """Load all publishers, and fill the cache"""
        if not self._has_all_publishers:
            self._publisher_cache = self._get_all_publishers()
            self._has_all_publishers = True

        return self._publisher_cache

    def get_scope(self, ref: str) -> Scope:
        id = ref.split("/")[-1]
        if (scope := self._scopes_cache.get(id)) is not None:
            return scope

        scope = self._get_scope(ref)
        self._scopes_cache[scope.id] = scope
        return scope

    def get_all_scopes(self) -> dict[str, Scope]:
        """Load all scopes, and fill the cache"""
        if not self._has_all_scopes:
            self._scopes_cache = self._get_all_scopes()
            self._has_all_scopes = True

        return self._scopes_cache


class _FileBasedSchemaLoader(CachedSchemaLoader):
    """Common logic for any schema loader that works with files (URLs or paths)"""

    def __init__(
        self,
        schema_url: URL | Path,
        *,
        loaded_callback: Callable[[DatasetSchema], None] | None = None,
    ):
        super().__init__()
        self.schema_url = schema_url
        self._loaded_callback = loaded_callback

    def __repr__(self):
        return f"{self.__class__.__name__}({self.schema_url!r})"

    @cached_property
    def _dataset_paths(self) -> dict[str, str]:
        """Cached index listing of all ID's to paths.
        This is typically needed for datasets that exist in sub folders,
        or when the folder name differs from the dataset ID.

        Since datasets can reference each other,
        this data is calculated for the whole repository at once.
        """
        return self._read_index()

    def _read_index(self) -> dict[str, str]:
        raise NotImplementedError

    def _read_dataset(self, dataset_id: str) -> Json:
        raise NotImplementedError

    def _read_view(self, dataset_id: str) -> str:
        raise NotImplementedError

    def _read_table(self, dataset_id: str, table_ref: str) -> Json:
        raise NotImplementedError

    def _get_dataset(self, dataset_id: str, prefetch_related: bool = False) -> DatasetSchema:
        """Gets a dataset from the filesystem for dataset_id."""
        schema_json = self._read_dataset(dataset_id)
        view_sql = self._read_view(dataset_id)
        return self._as_dataset(schema_json, view_sql, prefetch_related=prefetch_related)

    def _as_dataset(
        self, schema_json: dict, view_sql: str | None = None, prefetch_related: bool = False
    ) -> DatasetSchema:
        """Convert the read JSON into a real object that can resolve its relations."""
        dataset_schema = DatasetSchema(schema_json, view_sql, loader=self)

        if self._loaded_callback is not None:
            self._loaded_callback(dataset_schema)

        if prefetch_related:
            # Make sure the related datasets are read.
            for dataset_id in dataset_schema.related_dataset_schema_ids:
                if dataset_id != schema_json["id"]:  # skip self-references to local tables
                    self.get_dataset(dataset_id, prefetch_related=True)

        return dataset_schema

    def _get_dataset_path(self, dataset_id) -> str:
        """Find the relative path for a dataset."""
        try:
            # Since datasets are related, a cache is constructed.
            # Anything that is not part in this lookup table, will not be recognized as dataset.
            return self._dataset_paths[dataset_id]
        except KeyError:
            raise DatasetNotFound(
                f"Dataset '{dataset_id}' not found in '{self.schema_url}'."
            ) from None

    def _get_table(self, dataset: DatasetSchema, table_ref: str) -> DatasetTableSchema:
        """Load a versioned table from the location."""
        try:
            table_json = self._read_table(dataset.id, table_ref)
        except DatasetNotFound:
            # Amend the error for a better understanding. In this case, the dataset was
            # likely using a non-standard naming format (loaded via get_dataset_from_file())
            # but now the path isn't resolvable through the index. Fix the loading
            raise RuntimeError(f"Can't determine path to dataset '{dataset}'!") from None
        except SchemaObjectNotFound as e:
            # Provide better error message, can translate this into a more readable description
            raise DatasetTableNotFound(
                f"Dataset '{dataset.id}' has no table ref: '{table_ref}'!"
            ) from e
        return DatasetTableSchema(table_json, parent_schema=dataset)

    def _get_all_datasets(self) -> dict[str, DatasetSchema]:
        """Gets all datasets from the filesystem based on the `self.schema_url` path.
        Returns a dictionary of relative paths and their schema.
        """
        datasets = {}
        for dataset_id, dataset_path in sorted(self._dataset_paths.items()):
            dataset = self.get_dataset(dataset_id, prefetch_related=False)
            dataset.tables  # noqa: B018, ensure versioned tables are still prefetched
            datasets[dataset_path] = dataset
        return datasets

    def _get_publisher(self, publisher_id: str) -> Publisher:
        return Publisher.from_dict(
            read_json_path((self.root.parent / PUBLISHER_DIR / publisher_id).with_suffix(".json"))
        )

    def _get_all_publishers(self) -> dict[str, Publisher]:
        result = {}
        for path in (self.root.parent / PUBLISHER_DIR).glob("*.json"):
            if path.name in PUBLISHER_EXCLUDE_FILES:
                continue

            publisher = Publisher.from_dict(read_json_path(path))
            result[publisher.id] = publisher

        return result

    def _get_scope(self, ref: str) -> Scope:
        return Scope.from_dict(read_json_path(self.root.parent / f"{ref}.json"))

    def _get_all_scopes(self) -> dict[str, Scope]:
        result = {}
        for subdir in (self.root.parent / SCOPE_DIR).iterdir():
            for file in subdir.glob("*.json"):
                scope = Scope.from_dict(read_json_path(file))
                id = scope.db_name
                if id in result:
                    raise DuplicateScopeId(f'Scope ID "{id}" is already used in another scope')
                result[id] = scope
        return result


class FileSystemSchemaLoader(_FileBasedSchemaLoader):
    """Loader that loads dataset schemas from the filesystem."""

    def __init__(
        self,
        schema_url: Path | str,
        *,
        loaded_callback: Callable[[DatasetSchema], None] | None = None,
    ):
        """Initialize the loader with a folder where it needs to search for datasets.
        For the convenience of importing a selected subset, it's possible
        to point to a subfolder of the datasets repository.
        """
        schema_url = Path(schema_url) if isinstance(schema_url, str) else schema_url
        if not schema_url.exists():
            raise FileNotFoundError(schema_url)
        if not schema_url.is_dir():
            raise ValueError(
                f"FileSystemSchemaLoader should receive a folder, not a file: '{schema_url}'."
            )

        super().__init__(schema_url, loaded_callback=loaded_callback)
        try:
            # For compatibility with importing subfolders, this loader allows to define
            # a subfolder as target. To avoid unexpected content in the datasets database,
            # still try to calculate the real root so any calculated dataset-paths will work.
            self.root = self.get_root(schema_url)
        except ValueError:
            # In case the real root can't be found (typically in unit tests with random layouts),
            # assume the given folder should be treated as the root folder.
            self.root = schema_url

    @classmethod
    def from_file(cls, dataset_file: Path | str, **kwargs):
        """Helper function to support old patterns of loading random files as schema."""
        root = cls.get_root(dataset_file)
        return cls(schema_url=root, **kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.schema_url!r})"

    @classmethod
    def get_root(cls, dataset_file: Path | str) -> Path:
        """Resolve the real root folder.
        This makes sure that the datasets-path value will be correct,
        even when a subfolder is selected for importing files.
        """
        dataset_file = Path(dataset_file).resolve()
        dataset_file.stat()  # confirms the file exists, raises FileNotFoundError otherwise.
        if dataset_file.name == "datasets":
            # Pointed to the datasets folder already.
            return dataset_file

        if (try_root := dataset_file.joinpath("datasets")).exists():
            # Pointed to the root folder that has the 'datasets' folder.
            return try_root

        # Check whether it's a specific dataset inside the repository.
        # Find repository root from the file
        try:
            return next(dir for dir in dataset_file.parents if dir.name == "datasets")
        except StopIteration:
            raise ValueError(f"No 'datasets' root found for file '{dataset_file}'.") from None

    def get_dataset_from_file(
        self, dataset_file: Path | str, prefetch_related: bool = False, allow_external_files=False
    ):
        """Extra method, to read a dataset directly from a JSON file.
        This is mainly a helper function for testing.

        Normally, datasets are only detected when they use the format ``folder/dataset.json``.
        This method allows a more free-format naming convention for experimenting with files
        (e.g. useful for unit testing). It will however not be possible to resolve relations
        to other datasets when those datasets use same free-format for naming their files.

        Concluding, any relations will only resolve when:

        * the dataset follows the ``name/dataset.json`` convention;
        * or when the related dataset is also read and cached by the same loader instance.
        """
        # Cleanup the path
        dataset_file = Path(dataset_file)

        if not dataset_file.is_absolute():
            # Assume relative path from the repository root, typically used from unit test code.
            # For any imports that use this code path to import relative to the app directory,
            # an easy fix to is to convert to absolute paths first before calling this function.

            # Ensure that the path is lower than the root
            if "datasets" in (parts := dataset_file.parts):
                dataset_file = Path(*parts[parts.index("datasets") + 1 :])

            dataset_file = self.root.joinpath(dataset_file)
        dataset_file = dataset_file.resolve()  # removes ../../ entries, so is_relative_to() works

        if not allow_external_files and not dataset_file.is_relative_to(self.root):
            raise ValueError(
                f"Dataset file '{dataset_file}' does not exist in the schema repository"
            )

        schema_json = read_json_path(dataset_file)
        view_sql = _read_sql_path(dataset_file)
        return self._as_dataset(schema_json, view_sql, prefetch_related=prefetch_related)

    def _read_index(self) -> dict[str, str]:
        """A mapping of dataset ID to path."""
        # The index determines which datasets will be found.
        # For historical reasons, the filesystem loader can be initialized to work in a subfolder.
        # In that case, it will find fewer datasets, but still resolve them from the true root.
        id_to_path = {}
        for path in self.schema_url.glob("**/dataset.json"):
            file_json = read_json_path(path)
            if not isinstance(file_json, dict) or file_json.get("type") != "dataset":
                continue

            id_ = file_json.get("id")
            if id_ in id_to_path:
                raise RuntimeError(
                    f"Schema root '{self.root}' contains multiple datasets that named '{id_}', "
                    f"this will break relating datasets!"
                )
            id_to_path[id_] = str(path.parent.resolve().relative_to(self.root))
        return id_to_path

    def _read_dataset(self, dataset_id):
        dataset_path = self.get_dataset_path(dataset_id)
        return read_json_path(self.root / dataset_path / "dataset.json")

    def _read_table(self, dataset_id: str, table_ref: str) -> Json:
        dataset_path = self.get_dataset_path(dataset_id)
        return read_json_path(self.root / dataset_path / f"{table_ref}.json")

    def _read_view(self, dataset_id: str) -> str:
        dataset_path = self.get_dataset_path(dataset_id)
        return _read_sql_path(self.root / dataset_path / "dataset.sql")


def read_json_path(dataset_file: Path | str) -> Json:
    """Load JSON from a path."""
    dataset_file = Path(dataset_file)  # Path can take both string and Path
    try:
        with dataset_file.open() as stream:
            try:
                return json.load(stream)
            except json.JSONDecodeError as exc:
                raise ValueError("Invalid JSON file") from exc
    except FileNotFoundError as e:
        # Normalize the exception type for "not found" errors.
        raise SchemaObjectNotFound(str(dataset_file)) from e


def _read_sql_path(dataset_file: Path) -> str:
    """Load view SQL from a path"""
    try:
        with dataset_file.open() as stream:
            return stream.read()
    except FileNotFoundError:
        return None


def _read_sql_url(url: URL) -> str:
    """Load view SQL from an URL"""
    try:
        response = requests.get(url, timeout=10)
        if response.status_code != 200:
            # Normalize the exception type for "not found" errors.
            return None
        response.raise_for_status()  # All extend from OSError
        return response.text
    except requests.exceptions.ConnectionError:
        return None


class _SharedConnectionMixin:
    """Internal mixin for connection sharing."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._connection = None  # can be shared between methods

    @contextlib.contextmanager
    def _persistent_connection(self):
        """Context manager for having a single connection on retrieval."""
        if self._connection is not None:
            # Nested usage, just keep the connection.
            yield
        else:
            self._connection = requests.Session()
            with self._connection:  # calls .close() on exit
                yield
            self._connection = None

    def _read_json_url(self, url) -> Json:
        """Load JSON from an URL"""
        response = (self._connection or requests).get(url, timeout=60)
        if response.status_code == 404:
            # Normalize the exception type for "not found" errors.
            raise SchemaObjectNotFound(url)
        response.raise_for_status()  # All extend from OSError
        return response.json()


class URLSchemaLoader(_SharedConnectionMixin, _FileBasedSchemaLoader):
    """Loader that loads dataset schemas from an URL."""

    def __init__(
        self,
        schema_url: URL | str | None = None,
        *,
        loaded_callback: Callable[[DatasetSchema], None] | None = None,
    ):
        super().__init__(
            URL(schema_url or os.environ.get("SCHEMA_URL") or DEFAULT_SCHEMA_URL),
            loaded_callback=loaded_callback,
        )

    def _get_all_datasets(self) -> dict[str, DatasetSchema]:
        """Gets all datasets from a web url based on the `self.schema_url` path."""
        with self._persistent_connection():
            return super()._get_all_datasets()

    def _get_dataset(self, dataset_id: str, prefetch_related: bool = True) -> DatasetSchema:
        """Retrieve a dataset and its contents with a single connection."""
        with self._persistent_connection():
            return super()._get_dataset(dataset_id, prefetch_related=prefetch_related)

    def _read_index(self) -> dict[str, str]:
        return dict(self._read_json_url(self.schema_url / "index"))

    def _read_dataset(self, dataset_id: str) -> Json:
        dataset_path = self.get_dataset_path(dataset_id)
        return self._read_json_url(self.schema_url / dataset_path / "dataset")

    def _read_view(self, dataset_id: str) -> str:
        dataset_path = self.get_dataset_path(dataset_id)
        return _read_sql_url(self.schema_url / dataset_path / "dataset.sql")

    def _read_table(self, dataset_id: str, table_ref: str) -> Json:
        dataset_path = self.get_dataset_path(dataset_id)
        return self._read_json_url(self.schema_url / dataset_path / table_ref)

    def _get_publisher_url(self) -> URL:
        return URL(self.schema_url.rpartition("/datasets")[0]) / "publishers"

    def _get_scopes_url(self) -> URL:
        return URL(self.schema_url.rpartition("/datasets")[0]) / SCOPE_DIR

    def _get_publisher(self, publisher_id: str) -> Publisher:
        url = self._get_publisher_url()
        return Publisher.from_dict(self._read_json_url(url / publisher_id))

    def _get_all_publishers(self) -> dict[str, Publisher]:
        url = self._get_publisher_url()
        index = self._read_json_url(url / "index")
        result = {}
        for id_ in index:
            result[id_] = Publisher.from_dict(self._read_json_url(url / id_))

        return result

    def _get_scope(self, ref: str) -> Scope:
        base_url = URL(self.schema_url.rpartition("/datasets")[0])
        return Scope.from_dict(self._read_json_url(base_url / ref))

    def _get_all_scopes(self) -> dict[str, Scope]:
        url = self._get_scopes_url()
        index: dict[str, list[str]] = self._read_json_url(url / "index")
        result = {}
        for datateam, scope_list in index.items():
            for id_ in scope_list:
                if id_ in result:
                    raise DuplicateScopeId(f'Scope ID "{id_}" is already used in another scope')
                result[id_] = Scope.from_dict(self._read_json_url(url / datateam / id_))
        return result


class FileSystemProfileLoader(ProfileLoader):
    """Loading profiles from the file system."""

    def __init__(
        self,
        profiles_url: Path | str,
        *,
        loaded_callback: Callable[[ProfileSchema], None] | None = None,
    ):
        self.profiles_url = Path(profiles_url) if isinstance(profiles_url, str) else profiles_url
        self._loaded_callback = loaded_callback

    def get_profile(self, profile_id: str) -> ProfileSchema:
        """Load a specific profile by id."""
        data = read_json_path(self.profiles_url / f"{profile_id}.json")
        schema = ProfileSchema.from_dict(data)
        if self._loaded_callback is not None:
            self._loaded_callback(schema)
        return schema

    def get_all_profiles(self) -> list[ProfileSchema]:
        """Load all profiles found in a folder"""
        return [
            ProfileSchema.from_dict(read_json_path(path))
            for path in self.profiles_url.glob("**/*.json")
            if path.name != "index.json"
        ]


class URLProfileLoader(_SharedConnectionMixin, ProfileLoader):
    """Loading profiles from a URL"""

    def __init__(
        self,
        profiles_url: URL | str | None = None,
        *,
        loaded_callback: Callable[[ProfileSchema], None] | None = None,
    ):
        super().__init__()
        self.profiles_url = URL(
            profiles_url or os.environ.get("PROFILES_URL", DEFAULT_PROFILE_URL)
        )
        self._loaded_callback = loaded_callback

    def get_profile(self, datateam, profile_id: str) -> ProfileSchema:
        data = self._read_json_url(self.profiles_url / datateam / f"{profile_id}")
        schema = ProfileSchema.from_dict(data)
        if self._loaded_callback is not None:
            self._loaded_callback(schema)
        return schema

    def get_all_profiles(self) -> list[ProfileSchema]:
        profiles = []
        with self._persistent_connection():
            index: dict[str, list[str]] = self._read_json_url(self.profiles_url / "index")
            seen = []
            for datateam, profile_list in index.items():
                for name in profile_list:
                    profile = self.get_profile(datateam, name)
                    if profile.id in seen:
                        raise DuplicateProfileId(
                            f'Profile ID "{profile.id}" is already used in another profile'
                        )
                    seen.append(profile.id)
                    profiles.append(profile)
        return profiles


def get_schema_loader(schema_url: URL | str | None = None, **kwargs) -> CachedSchemaLoader:
    """Initialize the schema loader based on the given location.

    schema_url:
        Location where the schemas can be found. This
        can be a web url, or a filesystem path.
    """
    if schema_url is None:
        schema_url = os.environ.get("SCHEMA_URL") or DEFAULT_SCHEMA_URL

    if _is_url(schema_url):
        return URLSchemaLoader(schema_url, **kwargs)
    else:
        return FileSystemSchemaLoader(schema_url, **kwargs)


def get_profile_loader(profiles_url: URL | Path | str | None = None, **kwargs) -> ProfileLoader:
    """Initialize the profile loader for a given location."""
    if profiles_url is None:
        profiles_url = os.environ.get("PROFILES_URL") or DEFAULT_PROFILE_URL
    if _is_url(profiles_url):
        return URLProfileLoader(profiles_url, **kwargs)
    else:
        return FileSystemProfileLoader(profiles_url, **kwargs)


def _is_url(location: URL | str) -> bool:
    return isinstance(location, URL) or urlparse(location).scheme in ("http", "https")



from __future__ import annotations

import logging
import random
from collections import defaultdict
from typing import Any

from schematools.contrib.django.factories import DjangoModelFactory
from schematools.contrib.django.models import Dataset
from schematools.types import DatasetTableSchema

logger = logging.getLogger(__name__)


def add_temporal_attrs(
    field_name: str, value: Any, attrs: dict[str, Any], target_table: DatasetTableSchema
):
    """Add extra temporal attributes if target_table is temporal."""
    temporal_attrs = {}
    if temporal := target_table.temporal:
        temporal_identifier = temporal.identifier
        try:
            # The identificatie is one of the element of table.identifier
            identificatie = (set(target_table.identifier) - {temporal_identifier}).pop()
        except KeyError:
            logger.exception(
                "The temporal table `{target_table.name}` should have a composite key."
            )
        for postfix in (identificatie, temporal_identifier):
            try:
                postfix_value = getattr(value, postfix)
            except AttributeError:
                field_info = f"{target_table.name}:{field_name}"
                logger.warning(
                    "Skipping temporal attribute mocking for `%s` `%s`", field_info, postfix
                )
                return attrs
            temporal_attrs[field_name + "_" + postfix] = postfix_value
    return attrs | temporal_attrs


def relate_datasets(*datasets: Dataset) -> None:
    """Add relations to the datasets.

    There is one caveat. Because we are using the Django ORM,
    we do not have models for the extra through tabels for the 1-N models,
    that are added for the BenK related models.
    So these through tables cannot be filled with mock data.
    """
    models = defaultdict(dict)
    for dataset in datasets:
        if not dataset.enable_db:
            logger.warning("Skipping `%s`, `enable_db` is False", dataset.name)
            continue

        factory = DjangoModelFactory(dataset)
        for cls in factory.build_models():
            models[dataset.name][cls._meta.model_name] = cls

    for dataset in datasets:
        dataset_schema = dataset.schema
        for table in dataset_schema.tables:
            model = models[dataset_schema.db_name][
                table.db_name_variant(with_dataset_prefix=False, with_version=False)
            ]
            for f in table.fields:
                field_name = f.python_name

                if f.relation is not None:
                    # We need to get the related_model via the field on the source model
                    # For some reason, getting the related_model from the `models` dict
                    # does not work. The model class seems to be identical,
                    # however the check that Django applies is failing.
                    # Presumably because models are reloaded
                    # (Django warns about this during startup).
                    related_model = model._meta.get_field(field_name).remote_field.model

                    if isinstance(related_model, str):
                        logger.warning(
                            "Model `%s` cannot be resolved, relating will be skipped",
                            related_model,
                        )
                        continue

                    try:
                        values = list(related_model.objects.all())
                        nulled_objects = list(model.objects.all())
                    except ValueError as e:
                        message = str(e)
                        logger.warning("Values cannot be obtained: %s", message)
                        break

                    objs = []
                    if not values:
                        logger.warning("Skipping relate for %s, reason: no values", related_model)
                        continue
                    for obj in nulled_objects:
                        value = random.choice(values)  # noqa: S311 # nosec: B311
                        attrs = {field_name: value}
                        # Add the extra temporal fields that are stored separately
                        # on the object if applicable
                        if not f.is_loose_relation:
                            attrs = add_temporal_attrs(field_name, value, attrs, f.related_table)

                        for attr_name, attr_value in attrs.items():
                            setattr(
                                obj,
                                attr_name,
                                attr_value,
                            )
                        objs.append(obj)

                    if nulled_objects:
                        model.objects.bulk_update(objs, attrs.keys())

                elif f.nm_relation is not None:
                    target_model = model._meta.get_field(field_name).remote_field.model
                    if isinstance(target_model, str):
                        logger.warning(
                            "Skipping relate for %s, reason: model not loaded.", target_model
                        )
                        continue
                    target_values = target_model.objects.all()
                    source_values = model.objects.all()
                    m2m_descriptor = getattr(model, field_name)
                    through_model = m2m_descriptor.through

                    objs = []
                    source_id = table.db_name_variant(
                        with_dataset_prefix=False, with_version=False, postfix="_id"
                    )
                    target_id = field_name + "_id"

                    # first delete all relations
                    through_model.objects.all().delete()
                    for source_value in source_values:
                        target_value = random.choice(target_values)  # noqa: S311, # nosec B311
                        attrs = {source_id: source_value.pk, target_id: target_value.pk}
                        # Add the extra temporal fields that are stored separately
                        # on the object if applicable. This needs to be done for
                        # the source and target side of the relation.
                        if not f.is_loose_relation:
                            attrs = add_temporal_attrs(
                                table.db_name_variant(
                                    with_dataset_prefix=False, with_version=False
                                ),
                                source_value,
                                attrs,
                                table,
                            )
                            attrs = add_temporal_attrs(
                                field_name, target_value, attrs, f.related_table
                            )
                        objs.append(through_model.objects.create(**attrs))

                    if objs:
                        through_model.objects.bulk_update(objs, attrs.keys())



from __future__ import annotations

import os
from pathlib import Path

from psycopg import sql

from schematools.types import _PUBLIC_SCOPE, DatasetSchema, DatasetTableSchema


def export_geopackages(
    connection,
    dataset_schema: DatasetSchema,
    output: str,
    table_ids: list[str] | None = None,
    scopes: list[str] | None = None,
    size: int | None = None,
) -> None:
    """Export geopackages for all tables or an indicated subset in the dataset.

    Args:
        connection: SQLAlchemy connection object. Is needed for the SQL formatting.
        dataset_schema: Schema that needs export as geopackage.
        output: path on the filesystem where output should be stored.
        table_ids: optional parameter for a subset for the tables of the datasetself.
        scopes: Keycloak scopes that need to be taken into account.
            The geopackage will be produced contains information that is only
            accessible with these scopes.
        size: To produce a subset of the rows, mainly for testing.
    """

    base_dir = Path(output)

    pg_conn_str = (
        f"host={connection.engine.url.host} "
        f"port={connection.engine.url.port} "
        f"dbname={connection.engine.url.database} "
        f"user={connection.engine.url.username} "
        f"password={connection.engine.url.password}"
    )

    tables = (
        dataset_schema.tables
        if not table_ids
        else [dataset_schema.get_table_by_id(table_id) for table_id in table_ids]
    )

    for table in tables:
        # For now we only output the default version.
        output_path = base_dir / f"{table.db_name_variant(with_version=False)}.gpkg"
        field_names = sql.SQL(",").join(
            sql.Identifier(field.db_name)
            for field in _get_fields(dataset_schema, table, scopes or [])
            if field.db_name != "schema"
        )
        if not next(field_names.__iter__(), None):
            continue

        table_name = sql.Identifier(table.db_name)
        query = sql.SQL("SELECT {field_names} from {table_name}").format(
            field_names=field_names, table_name=table_name
        )
        if size is not None:
            query = sql.SQL("{query} LIMIT {size}").format(query=query, size=sql.Literal(size))

        with connection.connection.cursor() as cursor:
            sql_stmt = query.as_string(cursor)

        os.system(  # noqa: S605  # nosec: B605
            f'ogr2ogr -f "GPKG" {output_path} PG:"{pg_conn_str}" -sql "{sql_stmt}"'
        )


def _get_fields(dataset_schema: DatasetSchema, table: DatasetTableSchema, scopes: list[str]):
    parent_scopes = set(dataset_schema.auth | table.auth) - {_PUBLIC_SCOPE}
    for field in table.fields:
        if field.is_array:
            continue
        if field.is_internal:
            continue
        if parent_scopes | set(field.auth) - {_PUBLIC_SCOPE} <= set(scopes):
            yield field



from __future__ import annotations

from collections import deque
from collections.abc import Iterable
from pathlib import Path

from django.conf import settings
from django.core.management import BaseCommand, call_command
from django.db import transaction

from schematools import validation
from schematools.contrib.django.factories import DjangoModelFactory
from schematools.contrib.django.management.commands.migration_helpers import drop_table, migrate
from schematools.contrib.django.models import Dataset
from schematools.loaders import FileSystemSchemaLoader, get_schema_loader
from schematools.types import DatasetSchema, DatasetTableSchema

from .create_tables import create_tables
from .create_views import create_views


class Command(BaseCommand):
    help = """Import Amsterdam schema files.

    import_schemas imports all schemas from $SCHEMA_URL,
    unless schema files/URLs are given as positional arguments.
    """
    requires_system_checks = []

    def add_arguments(self, parser):
        parser.add_argument("schema", nargs="*", help="Local schema files to import")
        parser.add_argument(
            "--schema-url",
            default=settings.SCHEMA_URL,
            help=f"Schema URL (default: {settings.SCHEMA_URL})",
        )
        parser.add_argument("--create-tables", dest="create_tables", action="store_true")
        parser.add_argument("--create-views", dest="create_views", action="store_true")
        parser.add_argument("--no-create-tables", dest="create_tables", action="store_false")
        parser.add_argument("--dry-run", dest="dry_run", action="store_true")
        parser.add_argument("--execute", dest="dry_run", action="store_false")
        parser.add_argument("--migrate-tables", dest="migrate_tables", action="store_true")
        parser.add_argument("--no-migrate-tables", dest="migrate_tables", action="store_false")
        parser.set_defaults(create_tables=False)
        parser.set_defaults(create_views=False)
        parser.set_defaults(dry_run=True)
        parser.set_defaults(migrate_tables=True)

    def handle(self, *args, **options):
        self.dry_run = options["dry_run"]
        self.verbosity = options["verbosity"]
        self.schema_dependencies = deque()
        current_datasets = {
            Dataset.name_from_schema(ds.schema): ds for ds in Dataset.objects.all()
        }

        if options["schema"]:
            schemas = self.get_schemas_from_files(options["schema"])
        else:
            schemas = self.get_schemas_from_url(options["schema_url"])

        with transaction.atomic():
            if schemas and current_datasets:
                # Check if there are datasets missing
                missing_datasets = self.get_missing_datasets(current_datasets, schemas)

                if missing_datasets:
                    # If missing, delete these schemas from datasets_dataset table
                    for dataset in missing_datasets:
                        self.stdout.write(f"Missing dataset: {dataset.name}")
                        call_command("remove_schemas", dataset.name)

            # Contains unsaved Dataset objects if dry_run.
            updated_datasets = self._run_import(schemas)
            if not updated_datasets:
                self.stdout.write("No new datasets imported")
                return

            if options["migrate_tables"]:
                self._migrate_tables(current_datasets, updated_datasets, options)

            # Reasons for not creating tables directly are to manually configure the
            # "Datasets" model flags first. E.g. disable "enable_db".
            if options["create_tables"]:
                create_tables(self, updated_datasets, allow_unmanaged=True, dry_run=self.dry_run)

            if options["create_views"]:
                create_views(self, updated_datasets, dry_run=self.dry_run)

    def get_missing_datasets(
        self, current_datasets: dict[str, Dataset], schemas: dict[str, DatasetSchema]
    ) -> list[Dataset]:
        # Check if datasets_dataset table contains datasets not present in the schemas
        current_names = set(current_datasets.keys())

        # Get snake case names for schema objects
        schema_names = {Dataset.name_from_schema(schema_obj) for schema_obj in schemas.values()}

        # Datasets in current_datasets but missing in schemas
        missing_datasets = current_names - schema_names

        return [current_datasets[key] for key in list(missing_datasets) if key in current_datasets]

    def get_schemas_from_files(self, schema_files) -> dict[str, DatasetSchema]:
        """Import all schema definitions from the given files."""
        schemas = {}
        shared_loaders = {}

        def _get_shared_loader(path):
            # Make sure all datasets that share the same root also share the same loader,
            # so relations between the separate files can be resolved.
            if (loader := shared_loaders.get(path)) is None:
                shared_loaders[path] = loader = FileSystemSchemaLoader(
                    path, loaded_callback=self._loaded_callback
                )
            return loader

        for filename in schema_files:
            self.stdout.write(f"Loading schemas from {filename}")
            file = Path(filename).resolve()
            if file.is_dir():
                # As intended, read all datasets
                # The folder might be a sub path in the repository,
                # in which case only a few datasets will be imported.
                schemas.update(_get_shared_loader(file).get_all_datasets())
            else:
                # Previous logic also allowed selecting a single file by name.
                # This still needs to resolve the root to resolve relations,
                # and to make sure the 'path' is correctly calculated.
                root = FileSystemSchemaLoader.get_root(file)
                dataset = _get_shared_loader(root).get_dataset_from_file(file)
                # Random files may not follow the folder/dataset.json convention,
                # calculate the path here instead of using loader.get_dataset_path().
                path = file.parent.relative_to(root)
                if str(path) == ".":
                    path = dataset.id  # workaround for unit tests

                schemas.update({path: dataset})
        return schemas

    def get_schemas_from_url(self, schema_url) -> dict[str, DatasetSchema]:
        """Import all schema definitions from a URL"""
        self.stdout.write(f"Loading schema from {schema_url}")
        self.loader = get_schema_loader(schema_url, loaded_callback=self._loaded_callback)
        return self.loader.get_all_datasets()

    def _loaded_callback(self, schema: DatasetSchema):
        """Track which schema's get loaded. This is also used for dependency tracking."""
        self.stdout.write(f"* Loaded {schema.id}")

        self.schema_dependencies.append(schema)

    def _load_dependencies(self, dataset_schema: DatasetSchema, dataset) -> list[str]:
        """Make sure any dependencies are loaded.

        Returns the list of "real app names", which tells Django migrations those apps
        are not part of the project state, but can be found in the main app registry itself.
        """
        related_ids = dataset_schema.related_dataset_schema_ids
        real_apps = []

        # Load first, and this fills the cache.
        for dataset_id in related_ids - {dataset_schema.id}:
            dataset_schema.loader.get_dataset(dataset_id, prefetch_related=True)

        # Turn any loaded schema into a model.
        # And when a call to DjangoModelFactory.build_model() triggers loading of more schemas,
        # these are also picked up from the deque() collection object.
        while self.schema_dependencies:
            dependency_schema = self.schema_dependencies.popleft()
            if dependency_schema.id == dataset_schema.id:
                continue

            if self.verbosity >= 2:
                self.stdout.write(f"-- Building models for {dependency_schema.id}")
            DjangoModelFactory(dataset).build_models()
            real_apps.extend(
                [f"{dependency_schema.id}_{vmajor}" for vmajor in dependency_schema.versions]
            )

        return real_apps

    def _migrate_tables(
        self, current_datasets: dict[str, any], updated_datasets: Iterable[Dataset], options
    ):
        # Loop over updated datasets and perform migrations.
        for updated_dataset in updated_datasets:
            current_dataset = current_datasets.get(
                Dataset.name_from_schema(updated_dataset.schema)
            )
            if not current_dataset:  # New dataset
                continue

            real_apps = self._load_dependencies(updated_dataset.schema, updated_dataset)
            for current_table in current_dataset.schema.tables:
                updated_table = updated_dataset.schema.get_table_by_id(
                    current_table.id, include_nested=False, include_through=False
                )
                if current_table.version.vmajor == updated_table.version.vmajor:
                    # If the table is under_development and there are breaking changes to
                    # the table, drop the table
                    if current_table.status == DatasetTableSchema.Status.under_development:
                        previous_fields = current_table.json_data()["schema"]["properties"]
                        next_fields = updated_table.json_data()["schema"]["properties"]
                        table_errors = validation.validate_table(previous_fields, next_fields)
                        if len(table_errors) > 0:
                            if not options["create_tables"]:
                                self.stdout.write(
                                    "Not dropping table, as create_tables is set to false."
                                )
                            elif self.dry_run:
                                self.stdout.write(
                                    f"Would drop and replace table {current_table.db_name}."
                                )
                            else:
                                # drop the table and rely on create_tables to create it again.
                                for field in current_table.fields:
                                    if through_table := field.through_table:
                                        drop_table(through_table.db_name)
                                drop_table(current_table.db_name)
                            # do not migrate in this case.
                            continue

                    # Migrate the table, no breaking changes
                    migrate(
                        self,
                        current_dataset,
                        updated_dataset,
                        current_table,
                        updated_table,
                        real_apps,
                        dry_run=self.dry_run,
                    )

    def _run_import(self, dataset_schemas: dict[str, DatasetSchema]) -> list[Dataset]:
        datasets = []
        for id, schema in dataset_schemas.items():
            path = self.loader._get_dataset_path(id) if hasattr(self, "loader") else id
            self.stdout.write(f"* Processing {schema.id}")
            dataset = self._import(schema, path)
            if dataset is not None:
                datasets.append(dataset)

        return datasets

    def _import(self, schema: DatasetSchema, path: str) -> Dataset | None:
        """Import a single dataset schema."""
        try:
            dataset = Dataset.objects.get(name=Dataset.name_from_schema(schema))
            updated = dataset.save_for_schema(schema, path, save=not self.dry_run)
            if updated:
                self.stdout.write(f"  Updated {schema.id}")
                return dataset
        except Dataset.DoesNotExist:
            # Create new dataset
            dataset = Dataset.create_for_schema(schema, path, save=not self.dry_run)
            self.stdout.write(f"  Created {schema.id}")
            return dataset

        return None



from __future__ import annotations

from django.db import ProgrammingError, connection

from . import BaseDatasetCommand


class Command(BaseDatasetCommand):  # noqa: D101
    help = "Truncates the tables of a given dataset."
    requires_system_checks = []  # don't test URLs (which create models)

    def handle(self, *args, **options):  # noqa: D102
        db_tables = []
        for dataset in self.get_datasets(options, enable_db=True, default_all=False):
            db_tables.extend(
                table.db_name
                for table in dataset.schema.get_tables(include_nested=True, include_through=True)
            )

        with connection.cursor() as cursor:
            for db_table in sorted(db_tables):
                self.stdout.write(f"Truncating {db_table}")
                try:
                    cursor.execute(f"TRUNCATE {db_table}")
                except ProgrammingError:
                    # Catch missing tables, happens when views aren't generated on import_schemas
                    self.stdout.write(f"Failed to truncate {db_table}")



from __future__ import annotations

from django.conf import settings
from django.core.management import BaseCommand

from schematools.contrib.django.models import Publisher
from schematools.loaders import get_schema_loader
from schematools.types import Publisher as PublisherSchema


class Command(BaseCommand):
    help = "Import all known publishers from Amsterdam schema files."
    requires_system_checks = []

    def add_arguments(self, parser) -> None:
        parser.add_argument("publisher", nargs="*", help="Local publisher files to import")
        parser.add_argument(
            "--schema-url",
            default=settings.SCHEMA_URL,
            help=f"Schema URL (default: {settings.SCHEMA_URL})",
        )

    def handle(self, *args, **options) -> None:
        if options["publisher"]:
            publishers = self.import_from_files(options["publisher"])
        else:
            publishers = self.import_from_url(options["schema_url"])

        if publishers:
            self.stdout.write(f"Imported publishers: {len(publishers)}")
        else:
            self.stdout.write("No new publishers imported.")

    def import_from_files(self, publisher_files: list[str]) -> list[Publisher]:
        publishers = []
        for filename in publisher_files:
            schema = PublisherSchema.from_file(filename)
            publisher = self._import(schema)
            if publisher is not None:
                publishers.append(publisher)

        return publishers

    def import_from_url(self, schema_url: str) -> list[Publisher]:
        """Import all schema definitions from an URL"""
        self.stdout.write(f"Loading publishers from {schema_url}")
        loader = get_schema_loader(schema_url)
        publishers = []

        for _, schema in loader.get_all_publishers().items():
            publisher = self._import(schema)
            if publisher is not None:
                publishers.append(publisher)
        return publishers

    def _import(self, schema: PublisherSchema) -> Publisher | None:
        self.stdout.write(f"* Processing {schema.name}")
        id = schema.id
        try:
            publisher = Publisher.objects.get(id=id)
        except Publisher.DoesNotExist:
            publisher = Publisher.create_for_schema(schema)
            self.stdout.write(f"   Created {schema.name}")
            return publisher
        else:
            updated = publisher.save_for_schema(schema)
            if updated:
                self.stdout.write(f"    Updated {schema.name}")
                return publisher

        return None



from __future__ import annotations

from django.db import connection

from .factories import DjangoModelFactory
from .models import Dataset


def create_tables(dataset: Dataset):
    """Create the database tables for a given schema."""
    with connection.schema_editor() as schema_editor:
        factory = DjangoModelFactory(dataset)
        for model in factory.build_models():
            schema_editor.create_model(model)



# Generated by Django 4.1.3 on 2022-11-15 14:50
from __future__ import annotations

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0014_datasettable_version"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dataset",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="datasetfield",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="datasettable",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="profile",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
    ]



"""Very rudimentary implementation of fake geo data for now."""

from __future__ import annotations

import logging
import random

from django.contrib.gis import geos
from factory import Faker
from faker.providers import BaseProvider
from gisserver.geometries import CRS

from schematools import CRS_RD_NEW, SRID_3D, SRID_RD_NEW

LOCALE = "nl_NL"
DEFAULT_CRS = CRS_RD_NEW

# Amsterdam bounding box, in RD
WEST, SOUTH = 104000, 474000
EAST, NORTH = 136000, 501999

logger = logging.getLogger(__name__)


class UnsupportedGEOTypeException(Exception):
    """Exception to signal that the generic Geometry type is used.

    This type is not usable, because a faker does not know what type of geo data is needed.
    """


def random_point(has_z=False):
    point = (random.uniform(WEST, EAST), random.uniform(SOUTH, NORTH))  # noqa: S311  # nosec: B311
    return point if not has_z else point + (0.0,)


class Geometry:
    def __init__(self, has_z=False):
        class_name = self.__class__.__name__
        try:
            shape_cls = getattr(geos, class_name)
        except AttributeError:
            raise UnsupportedGEOTypeException(f"Class {class_name} is not allowed.") from None
        self.has_z = has_z
        self.shape = shape_cls(*self.get_coordinates(), srid=SRID_RD_NEW)

    def get_coordinates(self):
        return random_point(has_z=self.has_z)

    @property
    def ewkt(self):
        return self.shape.ewkt


class Point(Geometry):
    pass


class MultiPoint(Geometry):
    def get_coordinates(self):
        return [[random_point(has_z=self.has_z) for i in range(5)]]


class Polygon(Geometry):
    def get_coordinates(self):
        # Use a circle
        center = Point(has_z=self.has_z)
        return center.shape.buffer(1000)


class MultiPolygon(Geometry):
    def get_coordinates(self):
        return [[Polygon(has_z=self.has_z).shape for j in range(3)]]


class LineString(Geometry):
    def get_coordinates(self):
        return [[random_point(has_z=self.has_z) for i in range(5)]]


class MultiLineString(Geometry):
    def get_coordinates(self):
        return [[LineString(has_z=self.has_z).shape for j in range(3)]]


class GeometryCollection(Geometry):
    def get_coordinates(self):
        return [(Polygon(has_z=self.has_z).shape, LineString(has_z=self.has_z).shape)]


class GeoProvider(BaseProvider):  # noqa: D101
    """Geoprovider that selects a concrete implementation class."""

    def geo(self, geo_type: str = "Geometry", crs: CRS = DEFAULT_CRS) -> str:  # noqa: D102
        try:
            shape = globals()[geo_type](has_z=crs.srid in SRID_3D)
            if crs != DEFAULT_CRS:
                crs.apply_to(shape.shape)
            ewkt = shape.ewkt
        except (UnsupportedGEOTypeException, KeyError, TypeError) as e:
            message = str(e)
            logger.warning("Skipping geo data creation, reason: %s", message)
            return None
        else:
            return ewkt


Faker.add_provider(GeoProvider, locale=LOCALE)
